# /*##########################################################################
# Copyright (C) 2016-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""Tests for utils module"""

import io
import numpy
import os
import re
import shutil
import tempfile
import unittest

from .. import utils
from ..._version import calc_hexversion
import silx.io.url

import h5py
from ..utils import h5ls
from silx.io import commonh5


import fabio

__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "03/12/2020"

expected_spec1 = r"""#F .*
#D .*

#S 1 Ordinate1
#D .*
#N 2
#L Abscissa  Ordinate1
1  4\.00
2  5\.00
3  6\.00
"""

expected_spec2 = expected_spec1 + r"""
#S 2 Ordinate2
#D .*
#N 2
#L Abscissa  Ordinate2
1  7\.00
2  8\.00
3  9\.00
"""

expected_spec2reg = r"""#F .*
#D .*

#S 1 Ordinate1
#D .*
#N 3
#L Abscissa  Ordinate1  Ordinate2
1  4\.00  7\.00
2  5\.00  8\.00
3  6\.00  9\.00
"""

expected_spec2irr = expected_spec1 + r"""
#S 2 Ordinate2
#D .*
#N 2
#L Abscissa  Ordinate2
1  7\.00
2  8\.00
"""

expected_csv = r"""Abscissa;Ordinate1;Ordinate2
1;4\.00;7\.00e\+00
2;5\.00;8\.00e\+00
3;6\.00;9\.00e\+00
"""

expected_csv2 = r"""x;y0;y1
1;4\.00;7\.00e\+00
2;5\.00;8\.00e\+00
3;6\.00;9\.00e\+00
"""


class TestSave(unittest.TestCase):
    """Test saving curves as SpecFile:
    """

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.spec_fname = os.path.join(self.tempdir, "savespec.dat")
        self.csv_fname = os.path.join(self.tempdir, "savecsv.csv")
        self.npy_fname = os.path.join(self.tempdir, "savenpy.npy")

        self.x = [1, 2, 3]
        self.xlab = "Abscissa"
        self.y = [[4, 5, 6], [7, 8, 9]]
        self.y_irr = [[4, 5, 6], [7, 8]]
        self.ylabs = ["Ordinate1", "Ordinate2"]

    def tearDown(self):
        if os.path.isfile(self.spec_fname):
            os.unlink(self.spec_fname)
        if os.path.isfile(self.csv_fname):
            os.unlink(self.csv_fname)
        if os.path.isfile(self.npy_fname):
            os.unlink(self.npy_fname)
        shutil.rmtree(self.tempdir)

    def test_save_csv(self):
        utils.save1D(self.csv_fname, self.x, self.y,
                     xlabel=self.xlab, ylabels=self.ylabs,
                     filetype="csv", fmt=["%d", "%.2f", "%.2e"],
                     csvdelim=";", autoheader=True)

        csvf = open(self.csv_fname)
        actual_csv = csvf.read()
        csvf.close()

        self.assertRegex(actual_csv, expected_csv)

    def test_save_npy(self):
        """npy file is saved with numpy.save after building a numpy array
        and converting it to a named record array"""
        npyf = open(self.npy_fname, "wb")
        utils.save1D(npyf, self.x, self.y,
                     xlabel=self.xlab, ylabels=self.ylabs)
        npyf.close()

        npy_recarray = numpy.load(self.npy_fname)

        self.assertEqual(npy_recarray.shape, (3,))
        self.assertTrue(numpy.array_equal(npy_recarray['Ordinate1'],
                                          numpy.array((4, 5, 6))))

    def test_savespec_filename(self):
        """Save SpecFile using savespec()"""
        utils.savespec(self.spec_fname, self.x, self.y[0], xlabel=self.xlab,
                       ylabel=self.ylabs[0], fmt=["%d", "%.2f"],
                       close_file=True, scan_number=1)

        specf = open(self.spec_fname)
        actual_spec = specf.read()
        specf.close()
        self.assertRegex(actual_spec, expected_spec1)

    def test_savespec_file_handle(self):
        """Save SpecFile using savespec(), passing a file handle"""
        # first savespec: open, write file header, save y[0] as scan 1,
        #                 return file handle
        specf = utils.savespec(self.spec_fname, self.x, self.y[0],
                               xlabel=self.xlab, ylabel=self.ylabs[0],
                               fmt=["%d", "%.2f"], close_file=False)

        # second savespec: save y[1] as scan 2, close file
        utils.savespec(specf, self.x, self.y[1], xlabel=self.xlab,
                       ylabel=self.ylabs[1], fmt=["%d", "%.2f"],
                       write_file_header=False, close_file=True,
                       scan_number=2)

        specf = open(self.spec_fname)
        actual_spec = specf.read()
        specf.close()
        self.assertRegex(actual_spec, expected_spec2)

    def test_save_spec_reg(self):
        """Save SpecFile using save() on a regular pattern"""
        utils.save1D(self.spec_fname, self.x, self.y, xlabel=self.xlab,
                     ylabels=self.ylabs, filetype="spec", fmt=["%d", "%.2f"])

        specf = open(self.spec_fname)
        actual_spec = specf.read()
        specf.close()

        self.assertRegex(actual_spec, expected_spec2reg)

    def test_save_spec_irr(self):
        """Save SpecFile using save() on an irregular pattern"""
        # invalid  test case ?!
        return
        utils.save1D(self.spec_fname, self.x, self.y_irr, xlabel=self.xlab,
                     ylabels=self.ylabs, filetype="spec", fmt=["%d", "%.2f"])

        specf = open(self.spec_fname)
        actual_spec = specf.read()
        specf.close()
        self.assertRegex(actual_spec, expected_spec2irr)

    def test_save_csv_no_labels(self):
        """Save csv using save(), with autoheader=True but
        xlabel=None and ylabels=None
        This is a non-regression test for bug #223"""
        self.tempdir = tempfile.mkdtemp()
        self.spec_fname = os.path.join(self.tempdir, "savespec.dat")
        self.csv_fname = os.path.join(self.tempdir, "savecsv.csv")
        self.npy_fname = os.path.join(self.tempdir, "savenpy.npy")

        self.x = [1, 2, 3]
        self.xlab = "Abscissa"
        self.y = [[4, 5, 6], [7, 8, 9]]
        self.ylabs = ["Ordinate1", "Ordinate2"]
        utils.save1D(self.csv_fname, self.x, self.y,
                     autoheader=True, fmt=["%d", "%.2f", "%.2e"])

        csvf = open(self.csv_fname)
        actual_csv = csvf.read()
        csvf.close()
        self.assertRegex(actual_csv, expected_csv2)


def assert_match_any_string_in_list(test, pattern, list_of_strings):
    for string_ in list_of_strings:
        if re.match(pattern, string_):
            return True
    return False


class TestH5Ls(unittest.TestCase):
    """Test displaying the following HDF5 file structure:

        +foo
            +bar
                <HDF5 dataset "spam": shape (2, 2), type "<i8">
                <HDF5 dataset "tmp": shape (3,), type "<i8">
            <HDF5 dataset "data": shape (1,), type "<f8">

    """

    def assertMatchAnyStringInList(self, pattern, list_of_strings):
        for string_ in list_of_strings:
            if re.match(pattern, string_):
                return None
        raise AssertionError("regex pattern %s does not match any" % pattern +
                             " string in list " + str(list_of_strings))

    def testHdf5(self):
        fd, self.h5_fname = tempfile.mkstemp(text=False)
        # Close and delete (we just want the name)
        os.close(fd)
        os.unlink(self.h5_fname)
        self.h5f = h5py.File(self.h5_fname, "w")
        self.h5f["/foo/bar/tmp"] = [1, 2, 3]
        self.h5f["/foo/bar/spam"] = [[1, 2], [3, 4]]
        self.h5f["/foo/data"] = [3.14]
        self.h5f.close()

        rep = h5ls(self.h5_fname)
        lines = rep.split("\n")

        self.assertIn("+foo", lines)
        self.assertIn("\t+bar", lines)

        match = r'\t\t<HDF5 dataset "tmp": shape \(3,\), type "<i[48]">'
        self.assertMatchAnyStringInList(match, lines)
        match = r'\t\t<HDF5 dataset "spam": shape \(2, 2\), type "<i[48]">'
        self.assertMatchAnyStringInList(match, lines)
        match = r'\t<HDF5 dataset "data": shape \(1,\), type "<f[48]">'
        self.assertMatchAnyStringInList(match, lines)

        os.unlink(self.h5_fname)

    # Following test case disabled d/t errors on AppVeyor:
    #     os.unlink(spec_fname)
    # PermissionError: [WinError 32] The process cannot access the file because
    # it is being used by another process: 'C:\\...\\savespec.dat'

    # def testSpec(self):
    #     tempdir = tempfile.mkdtemp()
    #     spec_fname = os.path.join(tempdir, "savespec.dat")
    #
    #     x = [1, 2, 3]
    #     xlab = "Abscissa"
    #     y = [[4, 5, 6], [7, 8, 9]]
    #     ylabs = ["Ordinate1", "Ordinate2"]
    #     utils.save1D(spec_fname, x, y, xlabel=xlab,
    #                  ylabels=ylabs, filetype="spec",
    #                  fmt=["%d", "%.2f"])
    #
    #     rep = h5ls(spec_fname)
    #     lines = rep.split("\n")
    #     self.assertIn("+1.1", lines)
    #     self.assertIn("\t+instrument", lines)
    #
    #     self.assertMatchAnyStringInList(
    #             r'\t\t\t<SPEC dataset "file_header": shape \(\), type "|S60">',
    #             lines)
    #     self.assertMatchAnyStringInList(
    #             r'\t\t<SPEC dataset "Ordinate1": shape \(3L?,\), type "<f4">',
    #             lines)
    #
    #     os.unlink(spec_fname)
    #     shutil.rmtree(tempdir)


class TestOpen(unittest.TestCase):
    """Test `silx.io.utils.open` function."""

    @classmethod
    def setUpClass(cls):
        cls.tmp_directory = tempfile.mkdtemp()
        cls.createResources(cls.tmp_directory)

    @classmethod
    def createResources(cls, directory):

        cls.h5_filename = os.path.join(directory, "test.h5")
        h5 = h5py.File(cls.h5_filename, mode="w")
        h5["group/group/dataset"] = 50
        h5.close()

        cls.spec_filename = os.path.join(directory, "test.dat")
        utils.savespec(cls.spec_filename, [1], [1.1], xlabel="x", ylabel="y",
                       fmt=["%d", "%.2f"], close_file=True, scan_number=1)

        cls.edf_filename = os.path.join(directory, "test.edf")
        header = fabio.fabioimage.OrderedDict()
        header["integer"] = "10"
        data = numpy.array([[10, 50], [50, 10]])
        fabiofile = fabio.edfimage.EdfImage(data, header)
        fabiofile.write(cls.edf_filename)

        cls.txt_filename = os.path.join(directory, "test.txt")
        f = io.open(cls.txt_filename, "w+t")
        f.write(u"Kikoo")
        f.close()

        cls.missing_filename = os.path.join(directory, "test.missing")

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmp_directory)

    def testH5(self):
        f = utils.open(self.h5_filename)
        self.assertIsNotNone(f)
        self.assertIsInstance(f, h5py.File)
        f.close()

    def testH5With(self):
        with utils.open(self.h5_filename) as f:
            self.assertIsNotNone(f)
            self.assertIsInstance(f, h5py.File)

    def testH5_withPath(self):
        f = utils.open(self.h5_filename + "::/group/group/dataset")
        self.assertIsNotNone(f)
        self.assertEqual(f.h5py_class, h5py.Dataset)
        self.assertEqual(f[()], 50)
        f.close()

    def testH5With_withPath(self):
        with utils.open(self.h5_filename + "::/group/group") as f:
            self.assertIsNotNone(f)
            self.assertEqual(f.h5py_class, h5py.Group)
            self.assertIn("dataset", f)

    def testSpec(self):
        f = utils.open(self.spec_filename)
        self.assertIsNotNone(f)
        self.assertEqual(f.h5py_class, h5py.File)
        f.close()

    def testSpecWith(self):
        with utils.open(self.spec_filename) as f:
            self.assertIsNotNone(f)
            self.assertEqual(f.h5py_class, h5py.File)

    def testEdf(self):
        f = utils.open(self.edf_filename)
        self.assertIsNotNone(f)
        self.assertEqual(f.h5py_class, h5py.File)
        f.close()

    def testEdfWith(self):
        with utils.open(self.edf_filename) as f:
            self.assertIsNotNone(f)
            self.assertEqual(f.h5py_class, h5py.File)

    def testUnsupported(self):
        self.assertRaises(IOError, utils.open, self.txt_filename)

    def testNotExists(self):
        # load it
        self.assertRaises(IOError, utils.open, self.missing_filename)

    def test_silx_scheme(self):
        url = silx.io.url.DataUrl(scheme="silx", file_path=self.h5_filename, data_path="/")
        with utils.open(url.path()) as f:
            self.assertIsNotNone(f)
            self.assertTrue(silx.io.utils.is_file(f))

    def test_fabio_scheme(self):
        url = silx.io.url.DataUrl(scheme="fabio", file_path=self.edf_filename)
        self.assertRaises(IOError, utils.open, url.path())

    def test_bad_url(self):
        url = silx.io.url.DataUrl(scheme="sil", file_path=self.h5_filename)
        self.assertRaises(IOError, utils.open, url.path())

    def test_sliced_url(self):
        url = silx.io.url.DataUrl(file_path=self.h5_filename, data_slice=(5,))
        self.assertRaises(IOError, utils.open, url.path())


class TestNodes(unittest.TestCase):
    """Test `silx.io.utils.is_` functions."""

    def test_real_h5py_objects(self):
        name = tempfile.mktemp(suffix=".h5")
        try:
            with h5py.File(name, "w") as h5file:
                h5group = h5file.create_group("arrays")
                h5dataset = h5group.create_dataset("scalar", data=10)

                self.assertTrue(utils.is_file(h5file))
                self.assertTrue(utils.is_group(h5file))
                self.assertFalse(utils.is_dataset(h5file))

                self.assertFalse(utils.is_file(h5group))
                self.assertTrue(utils.is_group(h5group))
                self.assertFalse(utils.is_dataset(h5group))

                self.assertFalse(utils.is_file(h5dataset))
                self.assertFalse(utils.is_group(h5dataset))
                self.assertTrue(utils.is_dataset(h5dataset))
        finally:
            os.unlink(name)

    def test_h5py_like_file(self):

        class Foo(object):

            def __init__(self):
                self.h5_class = utils.H5Type.FILE

        obj = Foo()
        self.assertTrue(utils.is_file(obj))
        self.assertTrue(utils.is_group(obj))
        self.assertFalse(utils.is_dataset(obj))

    def test_h5py_like_group(self):

        class Foo(object):

            def __init__(self):
                self.h5_class = utils.H5Type.GROUP

        obj = Foo()
        self.assertFalse(utils.is_file(obj))
        self.assertTrue(utils.is_group(obj))
        self.assertFalse(utils.is_dataset(obj))

    def test_h5py_like_dataset(self):

        class Foo(object):

            def __init__(self):
                self.h5_class = utils.H5Type.DATASET

        obj = Foo()
        self.assertFalse(utils.is_file(obj))
        self.assertFalse(utils.is_group(obj))
        self.assertTrue(utils.is_dataset(obj))

    def test_bad(self):

        class Foo(object):

            def __init__(self):
                pass

        obj = Foo()
        self.assertFalse(utils.is_file(obj))
        self.assertFalse(utils.is_group(obj))
        self.assertFalse(utils.is_dataset(obj))

    def test_bad_api(self):

        class Foo(object):

            def __init__(self):
                self.h5_class = int

        obj = Foo()
        self.assertFalse(utils.is_file(obj))
        self.assertFalse(utils.is_group(obj))
        self.assertFalse(utils.is_dataset(obj))


class TestGetData(unittest.TestCase):
    """Test `silx.io.utils.get_data` function."""

    @classmethod
    def setUpClass(cls):
        cls.tmp_directory = tempfile.mkdtemp()
        cls.createResources(cls.tmp_directory)

    @classmethod
    def createResources(cls, directory):

        cls.h5_filename = os.path.join(directory, "test.h5")
        h5 = h5py.File(cls.h5_filename, mode="w")
        h5["group/group/scalar"] = 50
        h5["group/group/array"] = [1, 2, 3, 4, 5]
        h5["group/group/array2d"] = [[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]]
        h5.close()

        cls.spec_filename = os.path.join(directory, "test.dat")
        utils.savespec(cls.spec_filename, [1], [1.1], xlabel="x", ylabel="y",
                       fmt=["%d", "%.2f"], close_file=True, scan_number=1)

        cls.edf_filename = os.path.join(directory, "test.edf")
        cls.edf_multiframe_filename = os.path.join(directory, "test_multi.edf")
        header = fabio.fabioimage.OrderedDict()
        header["integer"] = "10"
        data = numpy.array([[10, 50], [50, 10]])
        fabiofile = fabio.edfimage.EdfImage(data, header)
        fabiofile.write(cls.edf_filename)
        fabiofile.append_frame(data=data, header=header)
        fabiofile.write(cls.edf_multiframe_filename)

        cls.txt_filename = os.path.join(directory, "test.txt")
        f = io.open(cls.txt_filename, "w+t")
        f.write(u"Kikoo")
        f.close()

        cls.missing_filename = os.path.join(directory, "test.missing")

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmp_directory)

    def test_hdf5_scalar(self):
        url = "silx:%s?/group/group/scalar" % self.h5_filename
        data = utils.get_data(url=url)
        self.assertEqual(data, 50)

    def test_hdf5_array(self):
        url = "silx:%s?/group/group/array" % self.h5_filename
        data = utils.get_data(url=url)
        self.assertEqual(data.shape, (5,))
        self.assertEqual(data[0], 1)

    def test_hdf5_array_slice(self):
        url = "silx:%s?path=/group/group/array2d&slice=1" % self.h5_filename
        data = utils.get_data(url=url)
        self.assertEqual(data.shape, (5,))
        self.assertEqual(data[0], 6)

    def test_hdf5_array_slice_out_of_range(self):
        url = "silx:%s?path=/group/group/array2d&slice=5" % self.h5_filename
        # ValueError: h5py 2.x
        # IndexError: h5py 3.x
        self.assertRaises((ValueError, IndexError), utils.get_data, url)

    def test_edf_using_silx(self):
        url = "silx:%s?/scan_0/instrument/detector_0/data" % self.edf_filename
        data = utils.get_data(url=url)
        self.assertEqual(data.shape, (2, 2))
        self.assertEqual(data[0, 0], 10)

    def test_fabio_frame(self):
        url = "fabio:%s?slice=1" % self.edf_multiframe_filename
        data = utils.get_data(url=url)
        self.assertEqual(data.shape, (2, 2))
        self.assertEqual(data[0, 0], 10)

    def test_fabio_singleframe(self):
        url = "fabio:%s?slice=0" % self.edf_filename
        data = utils.get_data(url=url)
        self.assertEqual(data.shape, (2, 2))
        self.assertEqual(data[0, 0], 10)

    def test_fabio_too_much_frames(self):
        url = "fabio:%s?slice=..." % self.edf_multiframe_filename
        self.assertRaises(ValueError, utils.get_data, url)

    def test_fabio_no_frame(self):
        url = "fabio:%s" % self.edf_filename
        data = utils.get_data(url=url)
        self.assertEqual(data.shape, (2, 2))
        self.assertEqual(data[0, 0], 10)

    def test_unsupported_scheme(self):
        url = "foo:/foo/bar"
        self.assertRaises(ValueError, utils.get_data, url)

    def test_no_scheme(self):
        url = "%s?path=/group/group/array2d&slice=5" % self.h5_filename
        self.assertRaises((ValueError, IOError), utils.get_data, url)

    def test_file_not_exists(self):
        url = "silx:/foo/bar"
        self.assertRaises(IOError, utils.get_data, url)


def _h5_py_version_older_than(version):
    v_majeur, v_mineur, v_micro = [int(i) for i in h5py.version.version.split('.')[:3]]
    r_majeur, r_mineur, r_micro = [int(i) for i in version.split('.')]
    return calc_hexversion(v_majeur, v_mineur, v_micro) >= calc_hexversion(r_majeur, r_mineur, r_micro) 


@unittest.skipUnless(_h5_py_version_older_than('2.9.0'), 'h5py version < 2.9.0')
class TestRawFileToH5(unittest.TestCase):
    """Test conversion of .vol file to .h5 external dataset"""

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self._vol_file = os.path.join(self.tempdir, 'test_vol.vol')
        self._file_info = os.path.join(self.tempdir, 'test_vol.info.vol')
        self._dataset_shape = 100, 20, 5
        data = numpy.random.random(self._dataset_shape[0] *
                                   self._dataset_shape[1] *
                                   self._dataset_shape[2]).astype(dtype=numpy.float32).reshape(self._dataset_shape)
        numpy.save(file=self._vol_file, arr=data)
        # those are storing into .noz file
        assert os.path.exists(self._vol_file + '.npy')
        os.rename(self._vol_file + '.npy', self._vol_file)
        self.h5_file = os.path.join(self.tempdir, 'test_h5.h5')
        self.external_dataset_path = '/root/my_external_dataset'
        self._data_url = silx.io.url.DataUrl(file_path=self.h5_file,
                                             data_path=self.external_dataset_path)
        with open(self._file_info, 'w') as _fi:
            _fi.write('NUM_X = %s\n' % self._dataset_shape[2])
            _fi.write('NUM_Y = %s\n' % self._dataset_shape[1])
            _fi.write('NUM_Z = %s\n' % self._dataset_shape[0])

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def check_dataset(self, h5_file, data_path, shape):
        """Make sure the external dataset is valid"""
        with h5py.File(h5_file, 'r') as _file:
            return data_path in _file and _file[data_path].shape == shape

    def test_h5_file_not_existing(self):
        """Test that can create a file with external dataset from scratch"""
        utils.rawfile_to_h5_external_dataset(bin_file=self._vol_file,
                                             output_url=self._data_url,
                                             shape=(100, 20, 5),
                                             dtype=numpy.float32)
        self.assertTrue(self.check_dataset(h5_file=self.h5_file,
                                           data_path=self.external_dataset_path,
                                           shape=self._dataset_shape))
        os.remove(self.h5_file)
        utils.vol_to_h5_external_dataset(vol_file=self._vol_file,
                                         output_url=self._data_url,
                                         info_file=self._file_info)
        self.assertTrue(self.check_dataset(h5_file=self.h5_file,
                                           data_path=self.external_dataset_path,
                                           shape=self._dataset_shape))

    def test_h5_file_existing(self):
        """Test that can add the external dataset from an existing file"""
        with h5py.File(self.h5_file, 'w') as _file:
            _file['/root/dataset1'] = numpy.zeros((100, 100))
            _file['/root/group/dataset2'] = numpy.ones((100, 100))
        utils.rawfile_to_h5_external_dataset(bin_file=self._vol_file,
                                             output_url=self._data_url,
                                             shape=(100, 20, 5),
                                             dtype=numpy.float32)
        self.assertTrue(self.check_dataset(h5_file=self.h5_file,
                                           data_path=self.external_dataset_path,
                                           shape=self._dataset_shape))

    def test_vol_file_not_existing(self):
        """Make sure error is raised if .vol file does not exists"""
        os.remove(self._vol_file)
        utils.rawfile_to_h5_external_dataset(bin_file=self._vol_file,
                                             output_url=self._data_url,
                                             shape=(100, 20, 5),
                                             dtype=numpy.float32)

        self.assertTrue(self.check_dataset(h5_file=self.h5_file,
                                           data_path=self.external_dataset_path,
                                           shape=self._dataset_shape))

    def test_conflicts(self):
        """Test several conflict cases"""
        # test if path already exists
        utils.rawfile_to_h5_external_dataset(bin_file=self._vol_file,
                                             output_url=self._data_url,
                                             shape=(100, 20, 5),
                                             dtype=numpy.float32)
        with self.assertRaises(ValueError):
            utils.rawfile_to_h5_external_dataset(bin_file=self._vol_file,
                                                 output_url=self._data_url,
                                                 shape=(100, 20, 5),
                                                 overwrite=False,
                                                 dtype=numpy.float32)

        utils.rawfile_to_h5_external_dataset(bin_file=self._vol_file,
                                             output_url=self._data_url,
                                             shape=(100, 20, 5),
                                             overwrite=True,
                                             dtype=numpy.float32)

        self.assertTrue(self.check_dataset(h5_file=self.h5_file,
                                           data_path=self.external_dataset_path,
                                           shape=self._dataset_shape))


class TestH5Strings(unittest.TestCase):
    """Test HDF5 str and bytes writing and reading"""

    @classmethod
    def setUpClass(cls):
        cls.tempdir = tempfile.mkdtemp()
        cls.vlenstr = h5py.special_dtype(vlen=str)
        cls.vlenbytes = h5py.special_dtype(vlen=bytes)
        try:
            cls.unicode = unicode
        except NameError:
            cls.unicode = str

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tempdir)

    def setUp(self):
        self.file = h5py.File(os.path.join(self.tempdir, 'file.h5'), mode="w")

    def tearDown(self):
        self.file.close()

    @classmethod
    def _make_array(cls, value, n):
        if isinstance(value, bytes):
            dtype = cls.vlenbytes
        elif isinstance(value, cls.unicode):
            dtype = cls.vlenstr
        else:
            return numpy.array([value] * n)
        return numpy.array([value] * n, dtype=dtype)

    @classmethod
    def _get_charset(cls, value):
        if isinstance(value, bytes):
            return h5py.h5t.CSET_ASCII
        elif isinstance(value, cls.unicode):
            return h5py.h5t.CSET_UTF8
        else:
            return None

    def _check_dataset(self, value, result=None):
        # Write+read scalar
        if result:
            decode_ascii = True
        else:
            decode_ascii = False
            result = value
        charset = self._get_charset(value)
        self.file["data"] = value
        data = utils.h5py_read_dataset(self.file["data"], decode_ascii=decode_ascii)
        assert type(data) == type(result), data
        assert data == result, data
        if charset:
            assert self.file["data"].id.get_type().get_cset() == charset

        # Write+read variable length
        self.file["vlen_data"] = self._make_array(value, 2)
        data = utils.h5py_read_dataset(self.file["vlen_data"], decode_ascii=decode_ascii, index=0)
        assert type(data) == type(result), data
        assert data == result, data
        data = utils.h5py_read_dataset(self.file["vlen_data"], decode_ascii=decode_ascii)
        numpy.testing.assert_array_equal(data, [result] * 2)
        if charset:
            assert self.file["vlen_data"].id.get_type().get_cset() == charset

    def _check_attribute(self, value, result=None):
        if result:
            decode_ascii = True
        else:
            decode_ascii = False
            result = value
        self.file.attrs["data"] = value
        data = utils.h5py_read_attribute(self.file.attrs, "data", decode_ascii=decode_ascii)
        assert type(data) == type(result), data
        assert data == result, data

        self.file.attrs["vlen_data"] = self._make_array(value, 2)
        data = utils.h5py_read_attribute(self.file.attrs, "vlen_data", decode_ascii=decode_ascii)
        assert type(data[0]) == type(result), data[0]
        assert data[0] == result, data[0]
        numpy.testing.assert_array_equal(data, [result] * 2)

        data = utils.h5py_read_attributes(self.file.attrs, decode_ascii=decode_ascii)["vlen_data"]
        assert type(data[0]) == type(result), data[0]
        assert data[0] == result, data[0]
        numpy.testing.assert_array_equal(data, [result] * 2)

    def test_dataset_ascii_bytes(self):
        self._check_dataset(b"abc")

    def test_attribute_ascii_bytes(self):
        self._check_attribute(b"abc")

    def test_dataset_ascii_bytes_decode(self):
        self._check_dataset(b"abc", result="abc")

    def test_attribute_ascii_bytes_decode(self):
        self._check_attribute(b"abc", result="abc")

    def test_dataset_ascii_str(self):
        self._check_dataset("abc")

    def test_attribute_ascii_str(self):
        self._check_attribute("abc")

    def test_dataset_utf8_str(self):
        self._check_dataset("\u0101bc")

    def test_attribute_utf8_str(self):
        self._check_attribute("\u0101bc")

    def test_dataset_utf8_bytes(self):
        # 0xC481 is the byte representation of U+0101
        self._check_dataset(b"\xc4\x81bc")

    def test_attribute_utf8_bytes(self):
        # 0xC481 is the byte representation of U+0101
        self._check_attribute(b"\xc4\x81bc")

    def test_dataset_utf8_bytes_decode(self):
        # 0xC481 is the byte representation of U+0101
        self._check_dataset(b"\xc4\x81bc", result="\u0101bc")

    def test_attribute_utf8_bytes_decode(self):
        # 0xC481 is the byte representation of U+0101
        self._check_attribute(b"\xc4\x81bc", result="\u0101bc")

    def test_dataset_latin1_bytes(self):
        # extended ascii character 0xE4
        self._check_dataset(b"\xe423")

    def test_attribute_latin1_bytes(self):
        # extended ascii character 0xE4
        self._check_attribute(b"\xe423")

    def test_dataset_latin1_bytes_decode(self):
        # U+DCE4: surrogate for extended ascii character 0xE4
        self._check_dataset(b"\xe423", result="\udce423")

    def test_attribute_latin1_bytes_decode(self):
        # U+DCE4: surrogate for extended ascii character 0xE4
        self._check_attribute(b"\xe423", result="\udce423")

    def test_dataset_no_string(self):
        self._check_dataset(numpy.int64(10))

    def test_attribute_no_string(self):
        self._check_attribute(numpy.int64(10))


def test_visitall_hdf5(tmp_path):
    """visit HDF5 file content not following links"""
    external_filepath = tmp_path / "external.h5"
    with h5py.File(external_filepath, mode="w") as h5file:
        h5file["target/dataset"] = 50

    filepath = tmp_path / "base.h5"
    with h5py.File(filepath, mode="w") as h5file:
        h5file["group/dataset"] = 50
        h5file["link/soft_link"] = h5py.SoftLink("/group/dataset")
        h5file["link/external_link"] = h5py.ExternalLink("external.h5", "/target/dataset")

    with h5py.File(filepath, mode="r") as h5file:
        visited_items = {}
        for path, item in utils.visitall(h5file):
            if isinstance(item, h5py.Dataset):
                content = item[()]
            elif isinstance(item, h5py.Group):
                content = None
            elif isinstance(item, h5py.SoftLink):
                content = item.path
            elif isinstance(item, h5py.ExternalLink):
                content = item.filename, item.path
            else:
                raise AssertionError("Item should not be present: %s" % path)
            visited_items[path] = (item.__class__, content)

    assert visited_items == {
        "/group": (h5py.Group, None),
        "/group/dataset": (h5py.Dataset, 50),
        "/link": (h5py.Group, None),
        "/link/soft_link": (h5py.SoftLink, "/group/dataset"),
        "/link/external_link": (h5py.ExternalLink, ("external.h5", "/target/dataset")),
    }

def test_visitall_commonh5():
    """Visit commonh5 File object"""
    fobj = commonh5.File("filename.file", mode="w")
    group = fobj.create_group("group")
    dataset = group.create_dataset("dataset", data=numpy.array(50))
    group["soft_link"] = dataset # Create softlink

    visited_items = dict(utils.visitall(fobj))
    assert len(visited_items) == 3
    assert visited_items["/group"] is group
    assert visited_items["/group/dataset"] is dataset
    soft_link = visited_items["/group/soft_link"]
    assert isinstance(soft_link, commonh5.SoftLink)
    assert soft_link.path == "/group/dataset"


def test_match_hdf5(tmp_path):
    """Test match function with HDF5 file"""
    with h5py.File(tmp_path / "test_match.h5", "w") as h5f:
        h5f.create_group("entry_0000/group")
        h5f["entry_0000/data"] = 0
        h5f.create_group("entry_0001/group")
        h5f["entry_0001/data"] = 1
        h5f.create_group("entry_0002")
        h5f["entry_0003"] = 3

        result = list(utils.match(h5f, "/entry_*/*"))

        assert sorted(result) == ['entry_0000/data', 'entry_0000/group', 'entry_0001/data', 'entry_0001/group']


def test_match_commonh5():
    """Test match function with commonh5 objects"""
    with commonh5.File("filename.file", mode="w") as fobj:
        fobj.create_group("entry_0000/group")
        fobj["entry_0000/data"] = 0
        fobj.create_group("entry_0001/group")
        fobj["entry_0001/data"] = 1
        fobj.create_group("entry_0002")
        fobj["entry_0003"] = 3

        result = list(utils.match(fobj, "/entry_*/*"))

        assert sorted(result) == ['entry_0000/data', 'entry_0000/group', 'entry_0001/data', 'entry_0001/group']



# /*##########################################################################
#
# Copyright (c) 2016-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["T. Vincent", "P. Knobel"]
__license__ = "MIT"
__date__ = "06/11/2018"


import numpy
import pytest

from silx.gui import qt
from silx.gui.colors import rgba
from silx.gui.colors import Colormap


@pytest.fixture(scope="module")
def sx(qapp):
    """Lazy loading to avoid it to create QApplication before qapp fixture"""
    from silx import sx
    if sx._IS_NOTEBOOK:
        pytest.skip("notebook context")
    if sx._NO_DISPLAY:
        pytest.skip("no DISPLAY specified")
    yield sx


def test_plot(sx, qapp_utils):
    """Test plot function"""
    y = numpy.random.random(100)
    x = numpy.arange(len(y)) * 0.5

    # Nothing
    plt = sx.plot()
    qapp_utils.exposeAndClose(plt)

    # y
    plt = sx.plot(y, title='y')
    qapp_utils.exposeAndClose(plt)

    # y, style
    plt = sx.plot(y, 'blued ', title='y, "blued "')
    qapp_utils.exposeAndClose(plt)

    # x, y
    plt = sx.plot(x, y, title='x, y')
    qapp_utils.exposeAndClose(plt)

    # x, y, style
    plt = sx.plot(x, y, 'ro-', xlabel='x', title='x, y, "ro-"')
    qapp_utils.exposeAndClose(plt)

    # x, y, style, y
    plt = sx.plot(x, y, 'ro-', y ** 2, xlabel='x', ylabel='y',
                  title='x, y, "ro-", y ** 2')
    qapp_utils.exposeAndClose(plt)

    # x, y, style, y, style
    plt = sx.plot(x, y, 'ro-', y ** 2, 'b--',
                  title='x, y, "ro-", y ** 2, "b--"')
    qapp_utils.exposeAndClose(plt)

    # x, y, style, x, y, style
    plt = sx.plot(x, y, 'ro-', x, y ** 2, 'b--',
                  title='x, y, "ro-", x, y ** 2, "b--"')
    qapp_utils.exposeAndClose(plt)

    # x, y, x, y
    plt = sx.plot(x, y, x, y ** 2, title='x, y, x, y ** 2')
    qapp_utils.exposeAndClose(plt)


def test_imshow(sx, qapp_utils):
    """Test imshow function"""
    img = numpy.arange(100.).reshape(10, 10) + 1

    # Nothing
    plt = sx.imshow()
    qapp_utils.exposeAndClose(plt)

    # image
    plt = sx.imshow(img)
    qapp_utils.exposeAndClose(plt)

    # image, named cmap
    plt = sx.imshow(img, cmap='jet', title='jet cmap')
    qapp_utils.exposeAndClose(plt)

    # image, custom colormap
    plt = sx.imshow(img, cmap=Colormap(), title='custom colormap')
    qapp_utils.exposeAndClose(plt)

    # image, log cmap
    plt = sx.imshow(img, norm='log', title='log cmap')
    qapp_utils.exposeAndClose(plt)

    # image, fixed range
    plt = sx.imshow(img, vmin=10, vmax=20,
                    title='[10,20] cmap')
    qapp_utils.exposeAndClose(plt)

    # image, keep ratio
    plt = sx.imshow(img, aspect=True,
                    title='keep ratio')
    qapp_utils.exposeAndClose(plt)

    # image, change origin and scale
    plt = sx.imshow(img, origin=(10, 10), scale=(2, 2),
                    title='origin=(10, 10), scale=(2, 2)')
    qapp_utils.exposeAndClose(plt)

    # image, origin='lower'
    plt = sx.imshow(img, origin='upper', title='origin="lower"')
    qapp_utils.exposeAndClose(plt)


def test_scatter(sx, qapp_utils):
    """Test scatter function"""
    x = numpy.arange(100)
    y = numpy.arange(100)
    values = numpy.arange(100)

    # simple scatter
    plt = sx.scatter(x, y, values)
    qapp_utils.exposeAndClose(plt)

    # No value
    plt = sx.scatter(x, y, values)
    qapp_utils.exposeAndClose(plt)

    # single value
    plt = sx.scatter(x, y, 10.)
    qapp_utils.exposeAndClose(plt)

    # set size
    plt = sx.scatter(x, y, values, size=20)
    qapp_utils.exposeAndClose(plt)

    # set colormap
    plt = sx.scatter(x, y, values, cmap='jet')
    qapp_utils.exposeAndClose(plt)

    # set colormap range
    plt = sx.scatter(x, y, values, vmin=2, vmax=50)
    qapp_utils.exposeAndClose(plt)

    # set colormap normalisation
    plt = sx.scatter(x, y, values, norm='log')
    qapp_utils.exposeAndClose(plt)


@pytest.mark.parametrize("plot_kind", ["plot", "imshow", "scatter"])
def test_ginput(sx, qapp, qapp_utils, plot_kind):
    """Test ginput function

    This does NOT perform interactive tests
    """
    create_plot = getattr(sx, plot_kind)
    plt = create_plot()
    qapp_utils.qWaitForWindowExposed(plt)
    qapp.processEvents()

    result = sx.ginput(1, timeout=0.1)
    assert len(result) == 0

    plt.setAttribute(qt.Qt.WA_DeleteOnClose)
    plt.close()


@pytest.mark.usefixtures("use_opengl")
def test_contour3d(sx, qapp_utils):
    """Test contour3d function"""
    coords = numpy.linspace(-10, 10, 64)
    z = coords.reshape(-1, 1, 1)
    y = coords.reshape(1, -1, 1)
    x = coords.reshape(1, 1, -1)
    data = numpy.sin(x * y * z) / (x * y * z)

    # Just data
    window = sx.contour3d(data)

    isosurfaces = window.getIsosurfaces()
    assert len(isosurfaces) == 1

    if not window.getPlot3DWidget().isValid():
        del window, isosurfaces  # Release widget reference
        pytest.skip("OpenGL context is not valid")

    # N contours + color
    colors = ['red', 'green', 'blue']
    window = sx.contour3d(data, copy=False, contours=len(colors),
                          color=colors)

    isosurfaces = window.getIsosurfaces()
    assert len(isosurfaces) == len(colors)
    for iso, color in zip(isosurfaces, colors):
        assert rgba(iso.getColor()) == rgba(color)

    # by isolevel, single color
    contours = 0.2, 0.5
    window = sx.contour3d(data, copy=False, contours=contours,
                          color='yellow')

    isosurfaces = window.getIsosurfaces()
    assert len(isosurfaces) == len(contours)
    for iso, level in zip(isosurfaces, contours):
        assert iso.getLevel() == level
        assert rgba(iso.getColor()) == rgba('yellow')

    # Single isolevel, colormap
    window = sx.contour3d(data, copy=False, contours=0.5,
                          colormap='gray', vmin=0.6, opacity=0.4)

    isosurfaces = window.getIsosurfaces()
    assert len(isosurfaces) == 1
    assert isosurfaces[0].getLevel() == 0.5
    assert rgba(isosurfaces[0].getColor()) == (0., 0., 0., 0.4)


@pytest.mark.usefixtures("use_opengl")
def test_points3d(sx, qapp_utils):
    """Test points3d function"""
    x = numpy.random.random(1024)
    y = numpy.random.random(1024)
    z = numpy.random.random(1024)
    values = numpy.random.random(1024)

    # 3D positions, no value
    window = sx.points3d(x, y, z)

    if not window.getSceneWidget().isValid():
        del window  # Release widget reference
        pytest.skip("OpenGL context is not valid")

    # 3D positions, values
    window = sx.points3d(x, y, z, values, mode='2dsquare',
                         colormap='magma', vmin=0.4, vmax=0.5)

    # 2D positions, no value
    window = sx.points3d(x, y)

    # 2D positions, values
    window = sx.points3d(x, y, values=values, mode=',',
                         colormap='magma', vmin=0.4, vmax=0.5)



# /*##########################################################################
# Copyright (C) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/

__authors__ = ["D. Naudet"]
__license__ = "MIT"
__date__ = "01/02/2016"

cimport numpy as cnumpy

cdef extern from "histogramnd_c.h":

    ctypedef enum histo_opt_type:
        HISTO_NONE
        HISTO_WEIGHT_MIN
        HISTO_WEIGHT_MAX
        HISTO_LAST_BIN_CLOSED

    ctypedef enum histo_rc_t:
        HISTO_OK
        HISTO_ERR_ALLOC

    # =====================
    # double sample, double cumul
    # =====================

    int histogramnd_double_double_double(double *i_sample,
                                         double *i_weigths,
                                         int i_n_dim,
                                         size_t i_n_elem,
                                         double *i_bin_ranges,
                                         int *i_n_bin,
                                         cnumpy.uint32_t *o_histo,
                                         double *o_cumul,
                                         double * bin_edges,
                                         int i_opt_flags,
                                         double i_weight_min,
                                         double i_weight_max) nogil

    int histogramnd_double_float_double(double *i_sample,
                                        float *i_weigths,
                                        int i_n_dim,
                                        size_t i_n_elem,
                                        double *i_bin_ranges,
                                        int *i_n_bin,
                                        cnumpy.uint32_t *o_histo,
                                        double *o_cumul,
                                        double * bin_edges,
                                        int i_opt_flags,
                                        float i_weight_min,
                                        float i_weight_max) nogil

    int histogramnd_double_int32_t_double(double *i_sample,
                                          cnumpy.int32_t *i_weigths,
                                          int i_n_dim,
                                          size_t i_n_elem,
                                          double *i_bin_ranges,
                                          int *i_n_bin,
                                          cnumpy.uint32_t *o_histo,
                                          double *o_cumul,
                                          double * bin_edges,
                                          int i_opt_flags,
                                          cnumpy.int32_t i_weight_min,
                                          cnumpy.int32_t i_weight_max) nogil

    # =====================
    # float sample, double cumul
    # =====================

    int histogramnd_float_double_double(float *i_sample,
                                        double *i_weigths,
                                        int i_n_dim,
                                        size_t i_n_elem,
                                        double *i_bin_ranges,
                                        int *i_n_bin,
                                        cnumpy.uint32_t *o_histo,
                                        double *o_cumul,
                                        double * bin_edges,
                                        int i_opt_flags,
                                        double i_weight_min,
                                        double i_weight_max) nogil

    int histogramnd_float_float_double(float *i_sample,
                                       float *i_weigths,
                                       int i_n_dim,
                                       size_t i_n_elem,
                                       double *i_bin_ranges,
                                       int *i_n_bin,
                                       cnumpy.uint32_t *o_histo,
                                       double *o_cumul,
                                       double * bin_edges,
                                       int i_opt_flags,
                                       float i_weight_min,
                                       float i_weight_max) nogil

    int histogramnd_float_int32_t_double(float *i_sample,
                                         cnumpy.int32_t *i_weigths,
                                         int i_n_dim,
                                         size_t i_n_elem,
                                         double *i_bin_ranges,
                                         int *i_n_bin,
                                         cnumpy.uint32_t *o_histo,
                                         double *o_cumul,
                                         double * bin_edges,
                                         int i_opt_flags,
                                         cnumpy.int32_t i_weight_min,
                                         cnumpy.int32_t i_weight_max) nogil

    # =====================
    # numpy.int32_t sample, double cumul
    # =====================

    int histogramnd_int32_t_double_double(cnumpy.int32_t *i_sample,
                                          double *i_weigths,
                                          int i_n_dim,
                                          size_t i_n_elem,
                                          double *i_bin_ranges,
                                          int *i_n_bin,
                                          cnumpy.uint32_t *o_histo,
                                          double *o_cumul,
                                          double * bin_edges,
                                          int i_opt_flags,
                                          double i_weight_min,
                                          double i_weight_max) nogil

    int histogramnd_int32_t_float_double(cnumpy.int32_t *i_sample,
                                         float *i_weigths,
                                         int i_n_dim,
                                         size_t i_n_elem,
                                         double *i_bin_ranges,
                                         int *i_n_bin,
                                         cnumpy.uint32_t *o_histo,
                                         double *o_cumul,
                                         double * bin_edges,
                                         int i_opt_flags,
                                         float i_weight_min,
                                         float i_weight_max) nogil

    int histogramnd_int32_t_int32_t_double(cnumpy.int32_t *i_sample,
                                           cnumpy.int32_t *i_weigths,
                                           int i_n_dim,
                                           size_t i_n_elem,
                                           double *i_bin_ranges,
                                           int *i_n_bin,
                                           cnumpy.uint32_t *o_histo,
                                           double *o_cumul,
                                           double * bin_edges,
                                           int i_opt_flags,
                                           cnumpy.int32_t i_weight_min,
                                           cnumpy.int32_t i_weight_max) nogil

    # =====================
    # double sample, float cumul
    # =====================

    int histogramnd_double_double_float(double *i_sample,
                                        double *i_weigths,
                                        int i_n_dim,
                                        size_t i_n_elem,
                                        double *i_bin_ranges,
                                        int *i_n_bin,
                                        cnumpy.uint32_t *o_histo,
                                        float *o_cumul,
                                        double * bin_edges,
                                        int i_opt_flags,
                                        double i_weight_min,
                                        double i_weight_max) nogil

    int histogramnd_double_float_float(double *i_sample,
                                       float *i_weigths,
                                       int i_n_dim,
                                       size_t i_n_elem,
                                       double *i_bin_ranges,
                                       int *i_n_bin,
                                       cnumpy.uint32_t *o_histo,
                                       float *o_cumul,
                                       double * bin_edges,
                                       int i_opt_flags,
                                       float i_weight_min,
                                       float i_weight_max) nogil

    int histogramnd_double_int32_t_float(double *i_sample,
                                         cnumpy.int32_t *i_weigths,
                                         int i_n_dim,
                                         size_t i_n_elem,
                                         double *i_bin_ranges,
                                         int *i_n_bin,
                                         cnumpy.uint32_t *o_histo,
                                         float *o_cumul,
                                         double * bin_edges,
                                         int i_opt_flags,
                                         cnumpy.int32_t i_weight_min,
                                         cnumpy.int32_t i_weight_max) nogil

    # =====================
    # float sample, float cumul
    # =====================

    int histogramnd_float_double_float(float *i_sample,
                                       double *i_weigths,
                                       int i_n_dim,
                                       size_t i_n_elem,
                                       double *i_bin_ranges,
                                       int *i_n_bin,
                                       cnumpy.uint32_t *o_histo,
                                       float *o_cumul,
                                       double * bin_edges,
                                       int i_opt_flags,
                                       double i_weight_min,
                                       double i_weight_max) nogil

    int histogramnd_float_float_float(float *i_sample,
                                      float *i_weigths,
                                      int i_n_dim,
                                      size_t i_n_elem,
                                      double *i_bin_ranges,
                                      int *i_n_bin,
                                      cnumpy.uint32_t *o_histo,
                                      float *o_cumul,
                                      double * bin_edges,
                                      int i_opt_flags,
                                      float i_weight_min,
                                      float i_weight_max) nogil

    int histogramnd_float_int32_t_float(float *i_sample,
                                        cnumpy.int32_t *i_weigths,
                                        int i_n_dim,
                                        size_t i_n_elem,
                                        double *i_bin_ranges,
                                        int *i_n_bin,
                                        cnumpy.uint32_t *o_histo,
                                        float *o_cumul,
                                        double * bin_edges,
                                        int i_opt_flags,
                                        cnumpy.int32_t i_weight_min,
                                        cnumpy.int32_t i_weight_max) nogil

    # =====================
    # numpy.int32_t sample, float cumul
    # =====================

    int histogramnd_int32_t_double_float(cnumpy.int32_t *i_sample,
                                         double *i_weigths,
                                         int i_n_dim,
                                         size_t i_n_elem,
                                         double *i_bin_ranges,
                                         int *i_n_bin,
                                         cnumpy.uint32_t *o_histo,
                                         float *o_cumul,
                                         double * bin_edges,
                                         int i_opt_flags,
                                         double i_weight_min,
                                         double i_weight_max) nogil

    int histogramnd_int32_t_float_float(cnumpy.int32_t *i_sample,
                                        float *i_weigths,
                                        int i_n_dim,
                                        size_t i_n_elem,
                                        double *i_bin_ranges,
                                        int *i_n_bin,
                                        cnumpy.uint32_t *o_histo,
                                        float *o_cumul,
                                        double * bin_edges,
                                        int i_opt_flags,
                                        float i_weight_min,
                                        float i_weight_max) nogil

    int histogramnd_int32_t_int32_t_float(cnumpy.int32_t *i_sample,
                                          cnumpy.int32_t *i_weigths,
                                          int i_n_dim,
                                          size_t i_n_elem,
                                          double *i_bin_ranges,
                                          int *i_n_bin,
                                          cnumpy.uint32_t *o_histo,
                                          float *o_cumul,
                                          double * bin_edges,
                                          int i_opt_flags,
                                          cnumpy.int32_t i_weight_min,
                                          cnumpy.int32_t i_weight_max) nogil



# /*##########################################################################
#
# Copyright (c) 2017-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides mix-in classes for :class:`Item3D`.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "24/04/2018"


import collections
import numpy

from silx.math.combo import min_max

from ...plot.items.core import ItemMixInBase
from ...plot.items.core import ColormapMixIn as _ColormapMixIn
from ...plot.items.core import SymbolMixIn as _SymbolMixIn
from ...plot.items.core import ComplexMixIn as _ComplexMixIn
from ...colors import rgba

from ..scene import primitives
from .core import Item3DChangedType, ItemChangedType


class InterpolationMixIn(ItemMixInBase):
    """Mix-in class for image interpolation mode

    :param str mode: 'linear' (default) or 'nearest'
    :param primitive:
        scene object for which to sync interpolation mode.
        This object MUST have an interpolation property that is updated.
    """

    NEAREST_INTERPOLATION = 'nearest'
    """Nearest interpolation mode (see :meth:`setInterpolation`)"""

    LINEAR_INTERPOLATION = 'linear'
    """Linear interpolation mode (see :meth:`setInterpolation`)"""

    INTERPOLATION_MODES = NEAREST_INTERPOLATION, LINEAR_INTERPOLATION
    """Supported interpolation modes for :meth:`setInterpolation`"""

    def __init__(self, mode=NEAREST_INTERPOLATION, primitive=None):
        self.__primitive = primitive
        self._syncPrimitiveInterpolation()

        self.__interpolationMode = None
        self.setInterpolation(mode)

    def _setPrimitive(self, primitive):

        """Set the scene object for which to sync interpolation"""
        self.__primitive = primitive
        self._syncPrimitiveInterpolation()

    def _syncPrimitiveInterpolation(self):
        """Synchronize scene object's interpolation"""
        if self.__primitive is not None:
            self.__primitive.interpolation = self.getInterpolation()

    def setInterpolation(self, mode):
        """Set image interpolation mode

        :param str mode: 'nearest' or 'linear'
        """
        mode = str(mode)
        assert mode in self.INTERPOLATION_MODES
        if mode != self.__interpolationMode:
            self.__interpolationMode = mode
            self._syncPrimitiveInterpolation()
            self._updated(Item3DChangedType.INTERPOLATION)

    def getInterpolation(self):
        """Returns the interpolation mode set by :meth:`setInterpolation`

        :rtype: str
        """
        return self.__interpolationMode


class ColormapMixIn(_ColormapMixIn):
    """Mix-in class for Item3D object with a colormap

    :param sceneColormap:
        The plot3d scene colormap to sync with Colormap object.
    """

    def __init__(self, sceneColormap=None):
        super(ColormapMixIn, self).__init__()

        self.__sceneColormap = sceneColormap
        self._syncSceneColormap()

    def _colormapChanged(self):
        """Handle colormap updates"""
        self._syncSceneColormap()
        super(ColormapMixIn, self)._colormapChanged()

    def _setSceneColormap(self, sceneColormap):
        """Set the scene colormap to sync with Colormap object.

        :param sceneColormap:
            The plot3d scene colormap to sync with Colormap object.
        """
        self.__sceneColormap = sceneColormap
        self._syncSceneColormap()

    def _getSceneColormap(self):
        """Returns scene colormap that is sync"""
        return self.__sceneColormap

    def _syncSceneColormap(self):
        """Synchronizes scene's colormap with Colormap object"""
        if self.__sceneColormap is not None:
            colormap = self.getColormap()

            self.__sceneColormap.colormap = colormap.getNColors()
            self.__sceneColormap.norm = colormap.getNormalization()
            self.__sceneColormap.gamma = colormap.getGammaNormalizationParameter()
            self.__sceneColormap.range_ = colormap.getColormapRange(self)
            self.__sceneColormap.nancolor = rgba(colormap.getNaNColor())


class ComplexMixIn(_ComplexMixIn):
    __doc__ = _ComplexMixIn.__doc__  # Reuse docstring

    _SUPPORTED_COMPLEX_MODES = (
        _ComplexMixIn.ComplexMode.REAL,
        _ComplexMixIn.ComplexMode.IMAGINARY,
        _ComplexMixIn.ComplexMode.ABSOLUTE,
        _ComplexMixIn.ComplexMode.PHASE,
        _ComplexMixIn.ComplexMode.SQUARE_AMPLITUDE)
    """Overrides supported ComplexMode"""


class SymbolMixIn(_SymbolMixIn):
    """Mix-in class for symbol and symbolSize properties for Item3D"""

    _SUPPORTED_SYMBOLS = collections.OrderedDict((
        ('o', 'Circle'),
        ('d', 'Diamond'),
        ('s', 'Square'),
        ('+', 'Plus'),
        ('x', 'Cross'),
        ('*', 'Star'),
        ('|', 'Vertical Line'),
        ('_', 'Horizontal Line'),
        ('.', 'Point'),
        (',', 'Pixel')))

    def _getSceneSymbol(self):
        """Returns a symbol name and size suitable for scene primitives.

        :return: (symbol, size)
        """
        symbol = self.getSymbol()
        size = self.getSymbolSize()
        if symbol == ',':  # pixel
            return 's', 1.
        elif symbol == '.':  # point
            # Size as in plot OpenGL backend, mimic matplotlib
            return 'o', numpy.ceil(0.5 * size) + 1.
        else:
            return symbol, size


class PlaneMixIn(ItemMixInBase):
    """Mix-in class for plane items (based on PlaneInGroup primitive)"""

    def __init__(self, plane):
        assert isinstance(plane, primitives.PlaneInGroup)
        self.__plane = plane
        self.__plane.alpha = 1.
        self.__plane.addListener(self._planeChanged)
        self.__plane.plane.addListener(self._planePositionChanged)

    def _getPlane(self):
        """Returns plane primitive

        :rtype: primitives.PlaneInGroup
        """
        return self.__plane

    def _planeChanged(self, source, *args, **kwargs):
        """Handle events from the plane primitive"""
        # Sync visibility
        if source.visible != self.isVisible():
            self.setVisible(source.visible)

    def _planePositionChanged(self, source, *args, **kwargs):
        """Handle update of cut plane position and normal"""
        if self.__plane.visible:  # TODO send even if hidden? or send also when showing if moved while hidden
            self._updated(ItemChangedType.POSITION)

    # Plane position

    def moveToCenter(self):
        """Move cut plane to center of data set"""
        self.__plane.moveToCenter()

    def isValid(self):
        """Returns whether the cut plane is defined or not (bool)"""
        return self.__plane.isValid

    def getNormal(self):
        """Returns the normal of the plane (as a unit vector)

        :return: Normal (nx, ny, nz), vector is 0 if no plane is defined
        :rtype: numpy.ndarray
        """
        return self.__plane.plane.normal

    def setNormal(self, normal):
        """Set the normal of the plane

        :param normal: 3-tuple of float: nx, ny, nz
        """
        self.__plane.plane.normal = normal

    def getPoint(self):
        """Returns a point on the plane

        :return: (x, y, z)
        :rtype: numpy.ndarray
        """
        return self.__plane.plane.point

    def setPoint(self, point):
        """Set a point contained in the plane.

        Warning: The plane might not intersect the bounding box of the data.

        :param point: (x, y, z) position
        :type point: 3-tuple of float
        """
        self.__plane.plane.point = point  # TODO rework according to PR #1303

    def getParameters(self):
        """Returns the plane equation parameters: a*x + b*y + c*z + d = 0

        :return: Plane equation parameters: (a, b, c, d)
        :rtype: numpy.ndarray
        """
        return self.__plane.plane.parameters

    def setParameters(self, parameters):
        """Set the plane equation parameters: a*x + b*y + c*z + d = 0

        Warning: The plane might not intersect the bounding box of the data.
        The given parameters will be normalized.

        :param parameters: (a, b, c, d) equation parameters
        """
        self.__plane.plane.parameters = parameters

    # Border stroke

    def _setForegroundColor(self, color):
        """Set the color of the plane border.

        :param color: RGBA color as 4 floats in [0, 1]
        """
        self.__plane.color = rgba(color)
        if hasattr(super(PlaneMixIn, self), '_setForegroundColor'):
            super(PlaneMixIn, self)._setForegroundColor(color)



# /*##########################################################################
# Copyright (C) 2017-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""Tests of the median filter"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "02/05/2017"

from silx.gui import qt
from silx.math.medianfilter import medfilt2d as medfilt2d_silx
import numpy
import numpy.random
from timeit import Timer
from silx.gui.plot import Plot1D
import logging

try:
    import scipy
except:
    scipy = None
else:
    import scipy.ndimage

try:
    import PyMca5.PyMca as pymca
except:
    pymca = None
else:
    from PyMca5.PyMca.median import medfilt2d as medfilt2d_pymca

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class BenchmarkMedianFilter(object):
    """Simple benchmark of the median fiter silx vs scipy"""

    NB_ITER = 3

    def __init__(self, imageWidth, kernels):
        self.img = numpy.random.rand(imageWidth, imageWidth)
        self.kernels = kernels

        self.run()

    def run(self):
        self.execTime = {}
        for kernel in self.kernels:
            self.execTime[kernel] = self.bench(kernel)

    def bench(self, width):
        def execSilx():
            medfilt2d_silx(self.img, width)

        def execScipy():
            scipy.ndimage.median_filter(input=self.img,
                                        size=width,
                                        mode='nearest')

        def execPymca():
            medfilt2d_pymca(self.img, width)

        execTime = {}

        t = Timer(execSilx)
        execTime["silx"] = t.timeit(BenchmarkMedianFilter.NB_ITER)
        logger.info(
            'exec time silx (kernel size = %s) is %s' % (width, execTime["silx"]))

        if scipy is not None:
            t = Timer(execScipy)
            execTime["scipy"] = t.timeit(BenchmarkMedianFilter.NB_ITER)
            logger.info(
                'exec time scipy (kernel size = %s) is %s' % (width, execTime["scipy"]))
        if pymca is not None:
            t = Timer(execPymca)
            execTime["pymca"] = t.timeit(BenchmarkMedianFilter.NB_ITER)
            logger.info(
                'exec time pymca (kernel size = %s) is %s' % (width, execTime["pymca"]))

        return execTime

    def getExecTimeFor(self, id):
        res = []
        for k in self.kernels:
            res.append(self.execTime[k][id])
        return res


app = qt.QApplication([])
kernels = [3, 5, 7, 11, 15]
benchmark = BenchmarkMedianFilter(imageWidth=1000, kernels=kernels)
plot = Plot1D()
plot.addCurve(x=kernels, y=benchmark.getExecTimeFor("silx"), legend='silx')
if scipy is not None:
    plot.addCurve(x=kernels, y=benchmark.getExecTimeFor("scipy"), legend='scipy')
if pymca is not None:
    plot.addCurve(x=kernels, y=benchmark.getExecTimeFor("pymca"), legend='pymca')
plot.show()
app.exec()
del app



# /*##########################################################################
#
# Copyright (c) 2017-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ###########################################################################*/
"""Test ScalarFieldView widget"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "17/01/2018"


import logging
import pytest

import numpy

from silx.utils.testutils import ParametricTestCase
from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt

from silx.gui.plot3d.ScalarFieldView import ScalarFieldView
from silx.gui.plot3d.SFViewParamTree import TreeView


_logger = logging.getLogger(__name__)


@pytest.mark.usefixtures("use_opengl")
class TestScalarFieldView(TestCaseQt, ParametricTestCase):
    """Tests of ScalarFieldView widget."""

    def setUp(self):
        super(TestScalarFieldView, self).setUp()
        self.widget = ScalarFieldView()
        self.widget.show()

        paramTreeWidget = TreeView()
        paramTreeWidget.setSfView(self.widget)

        dock = qt.QDockWidget()
        dock.setWidget(paramTreeWidget)
        self.widget.addDockWidget(qt.Qt.BottomDockWidgetArea, dock)

        # Commented as it slows down the tests
        # self.qWaitForWindowExposed(self.widget)

    def tearDown(self):
        self.qapp.processEvents()
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        del self.widget
        super(TestScalarFieldView, self).tearDown()

    @staticmethod
    def _buildData(size):
        """Make a 3D dataset"""
        coords = numpy.linspace(-10, 10, size)
        z = coords.reshape(-1, 1, 1)
        y = coords.reshape(1, -1, 1)
        x = coords.reshape(1, 1, -1)
        return numpy.sin(x * y * z) / (x * y * z)

    def testSimple(self):
        """Set the data and an isosurface"""
        data = self._buildData(size=32)

        self.widget.setData(data)
        self.widget.addIsosurface(0.5, (1., 0., 0., 0.5))
        self.widget.addIsosurface(0.7, qt.QColor('green'))
        self.qapp.processEvents()

    def testNotFinite(self):
        """Test with NaN and inf in data set"""

        # Some NaNs and inf
        data = self._buildData(size=32)
        data[8, :, :] = numpy.nan
        data[16, :, :] = numpy.inf
        data[24, :, :] = - numpy.inf

        self.widget.addIsosurface(0.5, 'red')
        self.widget.setData(data, copy=True)
        self.qapp.processEvents()
        self.widget.setData(None)

        # All NaNs or inf
        data = numpy.empty((4, 4, 4), dtype=numpy.float32)
        for value in (numpy.nan, numpy.inf):
            with self.subTest(value=str(value)):
                data[:] = value
                self.widget.setData(data, copy=True)
                self.qapp.processEvents()

    def testIsoSliderNormalization(self):
        """Test set TreeView with a different isoslider normalization"""
        data = self._buildData(size=32)

        self.widget.setData(data)
        self.widget.addIsosurface(0.5, (1., 0., 0., 0.5))
        self.widget.addIsosurface(0.7, qt.QColor('green'))
        self.qapp.processEvents()

        # Add a second TreeView
        paramTreeWidget = TreeView(self.widget)
        paramTreeWidget.setIsoLevelSliderNormalization('arcsinh')
        paramTreeWidget.setSfView(self.widget)

        dock = qt.QDockWidget()
        dock.setWidget(paramTreeWidget)
        self.widget.addDockWidget(qt.Qt.BottomDockWidgetArea, dock)



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""Tests for h5py utilities"""

__authors__ = ["W. de Nolf"]
__license__ = "MIT"
__date__ = "27/01/2020"


import unittest
import os
import sys
import time
import shutil
import logging
import tempfile
import multiprocessing
from contextlib import contextmanager

from .. import h5py_utils
from ...utils.retry import RetryError, RetryTimeoutError

IS_WINDOWS = sys.platform == "win32"
logger = logging.getLogger()


def _subprocess_context_main(queue, contextmgr, *args, **kw):
    try:
        with contextmgr(*args, **kw):
            queue.put(None)
            queue.get()
    except Exception:
        queue.put(None)
        raise


@contextmanager
def _subprocess_context(contextmgr, *args, **kw):
    print("\nSTART", os.getpid())
    timeout = kw.pop("timeout", 10)
    queue = multiprocessing.Queue(maxsize=1)
    p = multiprocessing.Process(
        target=_subprocess_context_main, args=(queue, contextmgr) + args, kwargs=kw
    )
    p.start()
    try:
        queue.get(timeout=timeout)
        yield
    finally:
        queue.put(None)
        p.join(timeout)
        print(" EXIT", os.getpid())


@contextmanager
def _open_context(filename, **kw):
    try:
        print(os.getpid(), "OPEN", filename, kw)
        with h5py_utils.File(filename, **kw) as f:
            if kw.get("mode") == "w":
                f["check"] = True
                f.flush()
            yield f
    except Exception:
        print(" ", os.getpid(), "FAILED", filename, kw)
        raise
    else:
        print(" ", os.getpid(), "CLOSED", filename, kw)


def _cause_segfault():
    import ctypes

    i = ctypes.c_char(b"a")
    j = ctypes.pointer(i)
    c = 0
    while True:
        j[c] = b"a"
        c += 1


def _top_level_names_test(txtfilename, *args, **kw):
    sys.stderr = open(os.devnull, "w")

    with open(txtfilename, mode="r") as f:
        failcounter = int(f.readline().strip())

    ncausefailure = kw.pop("ncausefailure")
    faildelay = kw.pop("faildelay")
    if failcounter < ncausefailure:
        time.sleep(faildelay)
        failcounter += 1
        with open(txtfilename, mode="w") as f:
            f.write(str(failcounter))
        if failcounter % 2:
            raise RetryError
        else:
            _cause_segfault()
    return h5py_utils._top_level_names(*args, **kw)


top_level_names_test = h5py_utils.retry_in_subprocess()(_top_level_names_test)


def subtests(test):
    def wrapper(self):
        for subtest_options in self._subtests():
            print("\n====SUB TEST===\n")
            print(f"sub test options: {subtest_options}")
            with self.subTest(str(subtest_options)):
                test(self)

    return wrapper


class TestH5pyUtils(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def _subtests(self):
        self._subtest_options = {"mode": "w"}
        self.filename_generator = self._filenames()
        yield self._subtest_options
        self._subtest_options = {"mode": "w", "libver": "latest"}
        self.filename_generator = self._filenames()
        yield

    def _filenames(self):
        i = 1
        while True:
            filename = os.path.join(self.test_dir, "file{}.h5".format(i))
            with self._open_context(filename):
                pass
            yield filename
            i += 1

    def _new_filename(self):
        return next(self.filename_generator)

    @contextmanager
    def _open_context(self, filename, **kwargs):
        kw = dict(self._subtest_options)
        kw.update(kwargs)
        with _open_context(filename, **kw) as f:
            yield f

    @contextmanager
    def _open_context_subprocess(self, filename, **kwargs):
        kw = dict(self._subtest_options)
        kw.update(kwargs)
        with _subprocess_context(_open_context, filename, **kw):
            yield

    def _assert_hdf5_data(self, f):
        self.assertTrue(f["check"][()])

    def _validate_hdf5_data(self, filename, swmr=False):
        with self._open_context(filename, mode="r") as f:
            self.assertEqual(f.swmr_mode, swmr)
            self._assert_hdf5_data(f)

    @subtests
    def test_modes_single_process(self):
        """Test concurrent access to the different files from the same process"""
        # When using HDF5_USE_FILE_LOCKING, open files with and without
        # locking should raise an exception. HDF5_USE_FILE_LOCKING should
        # be reset when all files are closed.

        orig = os.environ.get("HDF5_USE_FILE_LOCKING")
        filename1 = self._new_filename()
        self.assertEqual(orig, os.environ.get("HDF5_USE_FILE_LOCKING"))
        filename2 = self._new_filename()
        self.assertEqual(orig, os.environ.get("HDF5_USE_FILE_LOCKING"))

        with self._open_context(filename1, mode="r"):
            locking1 = False
            for mode in ["r", "w", "a"]:
                locking2 = mode != "r"
                raise_condition = not h5py_utils.HAS_LOCKING_ARGUMENT
                raise_condition &= locking1 != locking2
                with self.assertRaisesIf(raise_condition, RuntimeError):
                    with self._open_context(filename2, mode=mode):
                        pass
        self._validate_hdf5_data(filename1)
        self._validate_hdf5_data(filename2)
        self.assertEqual(orig, os.environ.get("HDF5_USE_FILE_LOCKING"))

        with self._open_context(filename1, mode="a"):
            locking1 = True
            for mode in ["r", "w", "a"]:
                locking2 = mode != "r"
                raise_condition = not h5py_utils.HAS_LOCKING_ARGUMENT
                raise_condition &= locking1 != locking2
                with self.assertRaisesIf(raise_condition, RuntimeError):
                    with self._open_context(filename2, mode=mode):
                        pass
        self._validate_hdf5_data(filename1)
        self._validate_hdf5_data(filename2)
        self.assertEqual(orig, os.environ.get("HDF5_USE_FILE_LOCKING"))

    @property
    def _libver_low_bound_is_v108(self):
        libver = self._subtest_options.get("libver")
        return h5py_utils._libver_low_bound_is_v108(libver)

    @property
    def _nonlocking_reader_before_writer(self):
        """A non-locking reader must open the file before it is locked by a writer"""
        if IS_WINDOWS and h5py_utils.HDF5_HAS_LOCKING_ARGUMENT:
            return True
        if not self._libver_low_bound_is_v108:
            return True
        return False

    @contextmanager
    def assertRaisesIf(self, condition, *args, **kw):
        if condition:
            with self.assertRaises(*args, **kw):
                yield
        else:
            yield

    @unittest.skipIf(
        h5py_utils.HDF5_HAS_LOCKING_ARGUMENT != h5py_utils.H5PY_HAS_LOCKING_ARGUMENT,
        "Versions of libhdf5 and h5py use incompatible file locking behaviour",
    )
    @subtests
    def test_modes_multi_process(self):
        """Test concurrent access to the same file from different processes"""
        filename = self._new_filename()

        nonlocking_reader_before_writer = self._nonlocking_reader_before_writer
        writer_before_nonlocking_reader_exception = OSError
        old_hdf5_on_windows = IS_WINDOWS and not h5py_utils.HDF5_HAS_LOCKING_ARGUMENT
        locked_exception = OSError

        # File locked by a writer
        unexpected_access = old_hdf5_on_windows and self._libver_low_bound_is_v108
        for wmode in ["w", "a"]:
            with self._open_context_subprocess(filename, mode=wmode):
                # Access by a second non-locking reader
                with self.assertRaisesIf(
                    nonlocking_reader_before_writer,
                    writer_before_nonlocking_reader_exception,
                ):
                    with self._open_context(filename, mode="r") as f:
                        self._assert_hdf5_data(f)
                # No access by a second locking reader
                if unexpected_access:
                    logger.warning("unexpected concurrent access by a locking reader")
                with self.assertRaisesIf(not unexpected_access, locked_exception):
                    with self._open_context(filename, mode="r", locking=True) as f:
                        self._assert_hdf5_data(f)
                # No access by a second writer
                if unexpected_access:
                    logger.warning("unexpected concurrent access by a writer")
                with self.assertRaisesIf(not unexpected_access, locked_exception):
                    with self._open_context(filename, mode="a") as f:
                        self._assert_hdf5_data(f)
                # Check for file corruption
                if not nonlocking_reader_before_writer:
                    self._validate_hdf5_data(filename)
            self._validate_hdf5_data(filename)

        # File locked by a reader
        unexpected_access = old_hdf5_on_windows
        with _subprocess_context(_open_context, filename, mode="r", locking=True):
            # Access by a non-locking reader
            with self._open_context(filename, mode="r") as f:
                self._assert_hdf5_data(f)
            # Access by a locking reader
            with self._open_context(filename, mode="r", locking=True) as f:
                self._assert_hdf5_data(f)
            # No access by a second writer
            if unexpected_access:
                logger.warning("unexpected concurrent access by a writer")
            raise_condition = not unexpected_access
            with self.assertRaisesIf(raise_condition, locked_exception):
                with self._open_context(filename, mode="a") as f:
                    self._assert_hdf5_data(f)
            # Check for file corruption
            self._validate_hdf5_data(filename)
        self._validate_hdf5_data(filename)

        # File open by a non-locking reader
        with self._open_context_subprocess(filename, mode="r"):
            # Access by a second non-locking reader
            with self._open_context(filename, mode="r") as f:
                self._assert_hdf5_data(f)
            # Access by a second locking reader
            with self._open_context(filename, mode="r", locking=True) as f:
                self._assert_hdf5_data(f)
            # Access by a second writer
            with self._open_context(filename, mode="a") as f:
                self._assert_hdf5_data(f)
            # Check for file corruption
            self._validate_hdf5_data(filename)
        self._validate_hdf5_data(filename)

    @subtests
    @unittest.skipIf(not h5py_utils.HAS_SWMR, "SWMR not supported")
    def test_modes_multi_process_swmr(self):
        filename = self._new_filename()

        with self._open_context(filename, mode="w", libver="latest") as f:
            pass

        # File open by SWMR writer
        with self._open_context_subprocess(filename, mode="a", swmr=True):
            with self._open_context(filename, mode="r") as f:
                assert f.swmr_mode
                self._assert_hdf5_data(f)
            with self.assertRaises(OSError):
                with self._open_context(filename, mode="a") as f:
                    pass
            self._validate_hdf5_data(filename, swmr=True)

    @subtests
    def test_retry_defaults(self):
        filename = self._new_filename()

        names = h5py_utils.top_level_names(filename)
        self.assertEqual(names, [])

        names = h5py_utils.safe_top_level_names(filename)
        self.assertEqual(names, [])

        names = h5py_utils.top_level_names(filename, include_only=None)
        self.assertEqual(names, ["check"])

        names = h5py_utils.safe_top_level_names(filename, include_only=None)
        self.assertEqual(names, ["check"])

        with h5py_utils.open_item(filename, "/check", validate=lambda x: False) as item:
            self.assertEqual(item, None)

        with h5py_utils.open_item(filename, "/check", validate=None) as item:
            self.assertTrue(item[()])

        with self.assertRaises(RetryTimeoutError):
            with h5py_utils.open_item(
                filename,
                "/check",
                retry_timeout=0.1,
                retry_invalid=True,
                validate=lambda x: False,
            ) as item:
                pass

        ncall = 0

        def validate(item):
            nonlocal ncall
            if ncall >= 1:
                return True
            else:
                ncall += 1
                raise RetryError

        with h5py_utils.open_item(
            filename,
            "/check",
            validate=validate,
            retry_timeout=1,
            retry_invalid=True,
        ) as item:
            self.assertTrue(item[()])

    @subtests
    def test_retry_custom(self):
        filename = self._new_filename()
        ncausefailure = 3
        faildelay = 0.1
        sufficient_timeout = ncausefailure * (faildelay + 10)
        insufficient_timeout = ncausefailure * faildelay * 0.5

        @h5py_utils.retry_contextmanager()
        def open_item(filename, name):
            nonlocal failcounter
            if failcounter < ncausefailure:
                time.sleep(faildelay)
                failcounter += 1
                raise RetryError
            with h5py_utils.File(filename) as h5file:
                yield h5file[name]

        failcounter = 0
        kw = {"retry_timeout": sufficient_timeout}
        with open_item(filename, "/check", **kw) as item:
            self.assertTrue(item[()])

        failcounter = 0
        kw = {"retry_timeout": insufficient_timeout}
        with self.assertRaises(RetryTimeoutError):
            with open_item(filename, "/check", **kw) as item:
                pass

    @subtests
    def test_retry_in_subprocess(self):
        filename = self._new_filename()
        txtfilename = os.path.join(self.test_dir, "failcounter.txt")
        ncausefailure = 3
        faildelay = 0.1
        sufficient_timeout = ncausefailure * (faildelay + 10)
        insufficient_timeout = ncausefailure * faildelay * 0.5

        kw = {
            "retry_timeout": sufficient_timeout,
            "include_only": None,
            "ncausefailure": ncausefailure,
            "faildelay": faildelay,
        }
        with open(txtfilename, mode="w") as f:
            f.write("0")
        names = top_level_names_test(txtfilename, filename, **kw)
        self.assertEqual(names, ["check"])

        kw = {
            "retry_timeout": insufficient_timeout,
            "include_only": None,
            "ncausefailure": ncausefailure,
            "faildelay": faildelay,
        }
        with open(txtfilename, mode="w") as f:
            f.write("0")
        with self.assertRaises(RetryTimeoutError):
            top_level_names_test(txtfilename, filename, **kw)

    @subtests
    def test_retry_generator(self):
        filename = self._new_filename()
        ncausefailure = 3
        faildelay = 0.1
        sufficient_timeout = ncausefailure * (faildelay + 10)
        insufficient_timeout = ncausefailure * faildelay * 0.5

        @h5py_utils.retry()
        def iter_data(filename, name, start_index=0):
            nonlocal failcounter
            if start_index <= 0:
                with h5py_utils.File(filename) as h5file:
                    yield h5file[name][()]
            if failcounter < ncausefailure:
                time.sleep(faildelay)
                failcounter += 1
                raise RetryError
            if start_index <= 1:
                with h5py_utils.File(filename) as h5file:
                    yield not h5file[name][()]

        failcounter = 0
        kw = {"retry_timeout": sufficient_timeout}
        data = list(iter_data(filename, "/check", **kw))
        self.assertEqual(data, [True, False])

        failcounter = 0
        kw = {"retry_timeout": insufficient_timeout}
        with self.assertRaises(RetryTimeoutError):
            list(iter_data(filename, "/check", **kw))



# /*##########################################################################
# Copyright (C) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
This module provides classes to calibrate data.

Classes
-------

- :class:`NoCalibration`
- :class:`LinearCalibration`
- :class:`ArrayCalibration`

"""
import numpy


class AbstractCalibration(object):
    """A calibration is a transformation to be applied to an axis (i.e. a 1D array).

    """
    def __init__(self):
        super(AbstractCalibration, self).__init__()

    def __call__(self, x):
        """Apply calibration to an axis or to a value.

        :param x: Axis (1-D array), or value"""
        raise NotImplementedError(
                "AbstractCalibration can not be used directly. " +
                "You must subclass it and implement __call__")

    def is_affine(self):
        """Returns True for an affine calibration of the form
        :math:`x  \\mapsto a + b * x`, or else False.
        """
        return False

    def get_slope(self):
        raise NotImplementedError(
                "get_slope is implemented only for affine calibrations")


class NoCalibration(AbstractCalibration):
    """No calibration :math:`x \\mapsto x`
    """
    def __init__(self):
        super(NoCalibration, self).__init__()

    def __call__(self, x):
        return x

    def is_affine(self):
        return True

    def get_slope(self):
        return 1.


class LinearCalibration(AbstractCalibration):
    """Linear calibration :math:`x \\mapsto a + b x`,
    where *a* is the y-intercept and *b* is the slope.

    :param y_intercept: y-intercept
    :param slope: Slope of the affine transformation
    """
    def __init__(self, y_intercept, slope):
        super(LinearCalibration, self).__init__()
        self.constant = y_intercept
        self.slope = slope

    def __call__(self, x):
        return self.constant + self.slope * x

    def is_affine(self):
        return True

    def get_slope(self):
        return self.slope


class ArrayCalibration(AbstractCalibration):
    """One-to-one mapping calibration, defined by an array *x'*,
    such as :math:`x \\mapsto x'`.

    This calibration can only be applied to x arrays of the same length as the
    calibration array *x'*.
    It is typically applied to an axis of indices or
    channels (:math:`0, 1, ..., n-1`).

    :param x1: Calibration array"""
    def __init__(self, x1):
        super(ArrayCalibration, self).__init__()
        if not isinstance(x1, (list, tuple)) and not hasattr(x1, "shape"):
            raise TypeError(
                    "The calibration array must be a sequence (list, dataset, array)")
        self.calibration_array = numpy.array(x1)
        self._is_affine = None

    def __call__(self, x):
        # calibrate the entire axis
        if isinstance(x, (list, tuple, numpy.ndarray)) and \
                        len(self.calibration_array) == len(x):
            return self.calibration_array
        # calibrate one value, by index
        if isinstance(x, int) and x < len(self.calibration_array):
            return self.calibration_array[x]
        raise ValueError("ArrayCalibration must be applied to array of same size "
                         "or to index.")

    def is_affine(self):
        """If all values in the calibration array are regularly spaced,
        return True."""
        if self._is_affine is None:
            delta_x = self.calibration_array[1:] - self.calibration_array[:-1]
            # use a less strict relative tolerance to account for rounding errors
            # e.g. when using float64 into float32 (see #1823)
            if not numpy.isclose(delta_x, delta_x[0], rtol=1e-4).all():
                self._is_affine = False
            else:
                self._is_affine = True
        return self._is_affine

    def get_slope(self):
        """If the calibration array is regularly spaced, return the spacing."""
        if not self.is_affine():
            raise AttributeError(
                "get_slope only makes sense for affine transformations"
            )
        return self.calibration_array[1] - self.calibration_array[0]


class FunctionCalibration(AbstractCalibration):
    """Calibration defined by a function *f*, such as :math:`x \\mapsto f(x)`*.

    :param function: Calibration function"""
    def __init__(self, function, is_affine=False):
        super(FunctionCalibration, self).__init__()
        if not hasattr(function, "__call__"):
            raise TypeError("The calibration function must be a callable")
        self.function = function
        self._is_affine = is_affine

    def __call__(self, x):
        return self.function(x)

    def is_affine(self):
        """Return True if calibration is affine.
        This is False by default, unless the object is instantiated with
        ``is_affine=True``."""
        return self._is_affine

    def get_slope(self):
        """If the calibration array is regularly spaced, return the spacing."""
        if not self.is_affine():
            raise AttributeError(
                "get_slope only makes sense for affine transformations"
            )
        # fixme: what if function is not defined at x=1 or x=2?
        return self.function(2) - self.function(1)



# /*##########################################################################
#
# Copyright (c) 2017-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Test suite for :class:`ComplexImageView`"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "17/01/2018"


import unittest
import logging
import numpy

from silx.utils.testutils import ParametricTestCase
from silx.gui.plot import ComplexImageView

from .utils import PlotWidgetTestCase


logger = logging.getLogger(__name__)


class TestComplexImageView(PlotWidgetTestCase, ParametricTestCase):
    """Test suite of ComplexImageView widget"""

    def _createPlot(self):
        return ComplexImageView.ComplexImageView()

    def testPlot2DComplex(self):
        """Test API of ComplexImageView widget"""
        data = numpy.array(((0, 1j), (1, 1 + 1j)), dtype=numpy.complex64)
        self.plot.setData(data)
        self.plot.setKeepDataAspectRatio(True)
        self.plot.getPlot().resetZoom()
        self.qWait(100)

        # Test colormap API
        colormap = self.plot.getColormap().copy()
        colormap.setName('magma')
        self.plot.setColormap(colormap)
        self.qWait(100)

        # Test all modes
        modes = self.plot.supportedComplexModes()
        for mode in modes:
            with self.subTest(mode=mode):
                self.plot.setComplexMode(mode)
                self.qWait(100)

        # Test origin and scale API
        self.plot.setScale((2, 1))
        self.qWait(100)
        self.plot.setOrigin((1, 1))
        self.qWait(100)

        # Test no data
        self.plot.setData(numpy.zeros((0, 0), dtype=numpy.complex64))
        self.qWait(100)

        # Test float data
        self.plot.setData(numpy.arange(100, dtype=numpy.float64).reshape(10, 10))
        self.qWait(100)



# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/
"""
Tests for the octaveh5 module
"""

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "12/07/2016"

import unittest
import os
import tempfile

try:
    from ..octaveh5 import Octaveh5
except ImportError:
    Octaveh5 = None


@unittest.skipIf(Octaveh5 is None, "Could not import h5py")
class TestOctaveH5(unittest.TestCase):
    @staticmethod
    def _get_struct_FT():
        return { 
            'NO_CHECK': 0.0, 'SHOWSLICE': 1.0, 'DOTOMO': 1.0, 'DATABASE': 0.0, 'ANGLE_OFFSET': 0.0, 
            'VOLSELECTION_REMEMBER': 0.0, 'NUM_PART': 4.0, 'VOLOUTFILE': 0.0, 'RINGSCORRECTION': 0.0, 
            'DO_TEST_SLICE': 1.0, 'ZEROOFFMASK': 1.0, 'VERSION': 'fastomo3 version 2.0', 
            'CORRECT_SPIKES_THRESHOLD': 0.040000000000000001, 'SHOWPROJ': 0.0, 'HALF_ACQ': 0.0, 
            'ANGLE_OFFSET_VALUE': 0.0, 'FIXEDSLICE': 'middle', 'VOLSELECT': 'total' }
    @staticmethod
    def _get_struct_PYHSTEXE():
        return {
            'EXE': 'PyHST2_2015d', 'VERBOSE': 0.0, 'OFFV': 'PyHST2_2015d', 'TOMO': 0.0, 
            'VERBOSE_FILE': 'pyhst_out.txt', 'DIR': '/usr/bin/', 'OFFN': 'pyhst2'}

    @staticmethod
    def _get_struct_FTAXIS():
        return {
            'POSITION_VALUE': 12345.0, 'COR_ERROR': 0.0, 'FILESDURINGSCAN': 0.0, 'PLOTFIGURE': 1.0,
            'DIM1': 0.0, 'OVERSAMPLING': 5.0, 'TO_THE_CENTER': 1.0, 'POSITION': 'fixed', 
            'COR_POSITION': 0.0, 'HA': 0.0 }
            
    @staticmethod
    def _get_struct_PAGANIN():
        return {
            'MKEEP_MASK': 0.0, 'UNSHARP_SIGMA': 0.80000000000000004, 'DILATE': 2.0, 'UNSHARP_COEFF': 3.0, 
            'MEDIANR': 4.0, 'DB': 500.0, 'MKEEP_ABS': 0.0, 'MODE': 0.0, 'THRESHOLD': 0.5, 
            'MKEEP_BONE': 0.0, 'DB2': 100.0, 'MKEEP_CORR': 0.0, 'MKEEP_SOFT': 0.0 }

    @staticmethod
    def _get_struct_BEAMGEO():
        return {'DIST': 55.0, 'SY': 0.0, 'SX': 0.0, 'TYPE': 'p'}


    def setUp(self):
        self.tempdir = tempfile.mkdtemp()        
        self.test_3_6_fname = os.path.join(self.tempdir, "silx_tmp_t00_octaveTest_3_6.h5")
        self.test_3_8_fname = os.path.join(self.tempdir, "silx_tmp_t00_octaveTest_3_8.h5")

    def tearDown(self):
        if os.path.isfile(self.test_3_6_fname):
            os.unlink(self.test_3_6_fname)
        if os.path.isfile(self.test_3_8_fname):
            os.unlink(self.test_3_8_fname)

    def testWritedIsReaded(self):
        """
        Simple test to write and reaf the structure compatible with the octave h5 using structure.
        This test is for # test for octave version > 3.8
        """   
        writer = Octaveh5()

        writer.open(self.test_3_8_fname, 'a')
        # step 1 writing the file
        writer.write('FT', self._get_struct_FT())
        writer.write('PYHSTEXE', self._get_struct_PYHSTEXE())
        writer.write('FTAXIS', self._get_struct_FTAXIS())
        writer.write('PAGANIN', self._get_struct_PAGANIN())
        writer.write('BEAMGEO', self._get_struct_BEAMGEO())
        writer.close()

        # step 2 reading the file
        reader = Octaveh5().open(self.test_3_8_fname)
        #  2.1 check FT
        data_readed = reader.get('FT')
        self.assertEqual(data_readed, self._get_struct_FT() )
        #  2.2 check PYHSTEXE
        data_readed = reader.get('PYHSTEXE')
        self.assertEqual(data_readed, self._get_struct_PYHSTEXE() )
        #  2.3 check FTAXIS
        data_readed = reader.get('FTAXIS')
        self.assertEqual(data_readed, self._get_struct_FTAXIS() )
        #  2.4 check PAGANIN
        data_readed = reader.get('PAGANIN')
        self.assertEqual(data_readed, self._get_struct_PAGANIN() )
        #  2.5 check BEAMGEO
        data_readed = reader.get('BEAMGEO')
        self.assertEqual(data_readed, self._get_struct_BEAMGEO() )
        reader.close()

    def testWritedIsReadedOldOctaveVersion(self):
        """The same test as testWritedIsReaded but for octave version < 3.8
        """
        # test for octave version < 3.8
        writer = Octaveh5(3.6)

        writer.open(self.test_3_6_fname, 'a')

        # step 1 writing the file
        writer.write('FT', self._get_struct_FT())
        writer.write('PYHSTEXE', self._get_struct_PYHSTEXE())
        writer.write('FTAXIS', self._get_struct_FTAXIS())
        writer.write('PAGANIN', self._get_struct_PAGANIN())
        writer.write('BEAMGEO', self._get_struct_BEAMGEO())
        writer.close()

        # step 2 reading the file
        reader = Octaveh5(3.6).open(self.test_3_6_fname)
        #  2.1 check FT
        data_readed = reader.get('FT')
        self.assertEqual(data_readed, self._get_struct_FT() )
        #  2.2 check PYHSTEXE
        data_readed = reader.get('PYHSTEXE')
        self.assertEqual(data_readed, self._get_struct_PYHSTEXE() )
        #  2.3 check FTAXIS
        data_readed = reader.get('FTAXIS')
        self.assertEqual(data_readed, self._get_struct_FTAXIS() )
        #  2.4 check PAGANIN
        data_readed = reader.get('PAGANIN')
        self.assertEqual(data_readed, self._get_struct_PAGANIN() )
        #  2.5 check BEAMGEO
        data_readed = reader.get('BEAMGEO')
        self.assertEqual(data_readed, self._get_struct_BEAMGEO() )
        reader.close()



# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
This module contains an :class:`DataFileDialog`.
"""

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "14/02/2018"

import enum
import logging
from silx.gui import qt
from silx.gui.hdf5.Hdf5Formatter import Hdf5Formatter
import silx.io
from .AbstractDataFileDialog import AbstractDataFileDialog

import fabio


_logger = logging.getLogger(__name__)


class _DataPreview(qt.QWidget):
    """Provide a preview of the selected image"""

    def __init__(self, parent=None):
        super(_DataPreview, self).__init__(parent)

        self.__formatter = Hdf5Formatter(self)
        self.__data = None
        self.__info = qt.QTableView(self)
        self.__model = qt.QStandardItemModel(self)
        self.__info.setModel(self.__model)
        self.__info.horizontalHeader().hide()
        self.__info.horizontalHeader().setStretchLastSection(True)
        layout = qt.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.__info)
        self.setLayout(layout)

    def colormap(self):
        return None

    def setColormap(self, colormap):
        # Ignored
        pass

    def sizeHint(self):
        return qt.QSize(200, 200)

    def setData(self, data, fromDataSelector=False):
        self.__info.setEnabled(data is not None)
        if data is None:
            self.__model.clear()
        else:
            self.__model.clear()

            if silx.io.is_dataset(data):
                kind = "Dataset"
            elif silx.io.is_group(data):
                kind = "Group"
            elif silx.io.is_file(data):
                kind = "File"
            else:
                kind = "Unknown"

            headers = []

            basename = data.name.split("/")[-1]
            if basename == "":
                basename = "/"
            headers.append("Basename")
            self.__model.appendRow([qt.QStandardItem(basename)])
            headers.append("Kind")
            self.__model.appendRow([qt.QStandardItem(kind)])
            if hasattr(data, "dtype"):
                headers.append("Type")
                text = self.__formatter.humanReadableType(data)
                self.__model.appendRow([qt.QStandardItem(text)])
            if hasattr(data, "shape"):
                headers.append("Shape")
                text = self.__formatter.humanReadableShape(data)
                self.__model.appendRow([qt.QStandardItem(text)])
            if hasattr(data, "attrs") and "NX_class" in data.attrs:
                headers.append("NX_class")
                value = data.attrs["NX_class"]
                formatter = self.__formatter.textFormatter()
                old = formatter.useQuoteForText()
                formatter.setUseQuoteForText(False)
                text = self.__formatter.textFormatter().toString(value)
                formatter.setUseQuoteForText(old)
                self.__model.appendRow([qt.QStandardItem(text)])
            self.__model.setVerticalHeaderLabels(headers)
        self.__data = data

    def __imageItem(self):
        image = self.__plot.getImage("data")
        return image

    def data(self):
        if self.__data is not None:
            if hasattr(self.__data, "name"):
                # in case of HDF5
                if self.__data.name is None:
                    # The dataset was closed
                    self.__data = None
        return self.__data

    def clear(self):
        self.__data = None
        self.__info.setText("")


class DataFileDialog(AbstractDataFileDialog):
    """The `DataFileDialog` class provides a dialog that allow users to select
    any datasets or groups from an HDF5-like file.

    The `DataFileDialog` class enables a user to traverse the file system in
    order to select an HDF5-like file. Then to traverse the file to select an
    HDF5 node.

    .. image:: img/datafiledialog.png

    The selected data is any kind of group or dataset. It can be restricted
    to only existing datasets or only existing groups using
    :meth:`setFilterMode`. A callback can be defining using
    :meth:`setFilterCallback` to filter even more data which can be returned.

    Filtering data which can be returned by a `DataFileDialog` can be done like
    that:

    .. code-block:: python

        # Force to return only a dataset
        dialog = DataFileDialog()
        dialog.setFilterMode(DataFileDialog.FilterMode.ExistingDataset)

    .. code-block:: python

        def customFilter(obj):
            if "NX_class" in obj.attrs:
                return obj.attrs["NX_class"] in [b"NXentry", u"NXentry"]
            return False

        # Force to return an NX entry
        dialog = DataFileDialog()
        # 1st, filter out everything which is not a group
        dialog.setFilterMode(DataFileDialog.FilterMode.ExistingGroup)
        # 2nd, check what NX_class is an NXentry
        dialog.setFilterCallback(customFilter)

    Executing a `DataFileDialog` can be done like that:

    .. code-block:: python

        dialog = DataFileDialog()
        result = dialog.exec()
        if result:
            print("Selection:")
            print(dialog.selectedFile())
            print(dialog.selectedUrl())
        else:
            print("Nothing selected")

    If the selection is a dataset you can access to the data using
    :meth:`selectedData`.

    If the selection is a group or if you want to read the selected object on
    your own you can use the `silx.io` API.

    .. code-block:: python

        url = dialog.selectedUrl()
        with silx.io.open(url) as data:
            pass

    Or by loading the file first

    .. code-block:: python

        url = dialog.selectedDataUrl()
        with silx.io.open(url.file_path()) as h5:
            data = h5[url.data_path()]

    Or by using `h5py` library

    .. code-block:: python

        url = dialog.selectedDataUrl()
        with h5py.File(url.file_path(), mode="r") as h5:
            data = h5[url.data_path()]
    """

    class FilterMode(enum.Enum):
        """This enum is used to indicate what the user may select in the
        dialog; i.e. what the dialog will return if the user clicks OK."""

        AnyNode = 0
        """Any existing node from an HDF5-like file."""
        ExistingDataset = 1
        """An existing HDF5-like dataset."""
        ExistingGroup = 2
        """An existing HDF5-like group. A file root is a group."""

    def __init__(self, parent=None):
        AbstractDataFileDialog.__init__(self, parent=parent)
        self.__filter = DataFileDialog.FilterMode.AnyNode
        self.__filterCallback = None

    def selectedData(self):
        """Returns the selected data by using the :meth:`silx.io.get_data`
        API with the selected URL provided by the dialog.

        If the URL identify a group of a file it will raise an exception. For
        group or file you have to use on your own the API :meth:`silx.io.open`.

        :rtype: numpy.ndarray
        :raise ValueError: If the URL do not link to a dataset
        """
        url = self.selectedUrl()
        return silx.io.get_data(url)

    def _createPreviewWidget(self, parent):
        previewWidget = _DataPreview(parent)
        previewWidget.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        return previewWidget

    def _createSelectorWidget(self, parent):
        # There is no selector
        return None

    def _createPreviewToolbar(self, parent, dataPreviewWidget, dataSelectorWidget):
        # There is no toolbar
        return None

    def _isDataSupportable(self, data):
        """Check if the selected data can be supported at one point.

        If true, the data selector will be checked and it will update the data
        preview. Else the selecting is disabled.

        :rtype: bool
        """
        # Everything is supported
        return True

    def _isFabioFilesSupported(self):
        # Everything is supported
        return False

    def _isDataSupported(self, data):
        """Check if the data can be returned by the dialog.

        If true, this data can be returned by the dialog and the open button
        will be enabled. If false the button will be disabled.

        :rtype: bool
        """
        if self.__filter == DataFileDialog.FilterMode.AnyNode:
            accepted = True
        elif self.__filter == DataFileDialog.FilterMode.ExistingDataset:
            accepted = silx.io.is_dataset(data)
        elif self.__filter == DataFileDialog.FilterMode.ExistingGroup:
            accepted = silx.io.is_group(data)
        else:
            raise ValueError("Filter %s is not supported" % self.__filter)
        if not accepted:
            return False
        if self.__filterCallback is not None:
            try:
                return self.__filterCallback(data)
            except Exception:
                _logger.error("Error while executing custom callback", exc_info=True)
                return False
        return True

    def setFilterCallback(self, callback):
        """Set the filter callback. This filter is applied only if the filter
        mode (:meth:`filterMode`) first accepts the selected data.

        It is not supposed to be set while the dialog is being used.

        :param callable callback: Define a custom function returning a boolean
            and taking as argument an h5-like node. If the function returns true
            the dialog can return the associated URL.
        """
        self.__filterCallback = callback

    def setFilterMode(self, mode):
        """Set the filter mode.

        It is not supposed to be set while the dialog is being used.

        :param DataFileDialog.FilterMode mode: The new filter.
        """
        self.__filter = mode

    def fileMode(self):
        """Returns the filter mode.

        :rtype: DataFileDialog.FilterMode
        """
        return self.__filter

    def _displayedDataInfo(self, dataBeforeSelection, dataAfterSelection):
        """Returns the text displayed under the data preview.

        This zone is used to display error in case or problem of data selection
        or problems with IO.

        :param numpy.ndarray dataAfterSelection: Data as it is after the
            selection widget (basically the data from the preview widget)
        :param numpy.ndarray dataAfterSelection: Data as it is before the
            selection widget (basically the data from the browsing widget)
        :rtype: bool
        """
        return u""



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

"""
Test of the Convolution class.
"""

__authors__ = ["Pierre Paleo"]
__contact__ = "pierre.paleo@esrf.fr"
__license__ = "MIT"
__copyright__ = "2019 European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "01/08/2019"

import pytest
import logging
from itertools import product
import numpy as np
from silx.image.utils import gaussian_kernel

try:
    from scipy.ndimage import convolve, convolve1d
    from scipy.misc import ascent

    scipy_convolve = convolve
    scipy_convolve1d = convolve1d
except ImportError:
    scipy_convolve = None
import unittest
from ..common import ocl, check_textures_availability

if ocl:
    import pyopencl as cl
    import pyopencl.array as parray
    from silx.opencl.convolution import Convolution
logger = logging.getLogger(__name__)


class ConvolutionData:

    def __init__(self, param):
        self.param = param
        self.mode = param["boundary_handling"]
        logger.debug(
            """
            Testing convolution with boundary_handling=%s,
            use_textures=%s, input_device=%s, output_device=%s
            """
            % (
                self.mode,
                param["use_textures"],
                param["input_on_device"],
                param["output_on_device"],
            )
        )

    @classmethod
    def setUpClass(cls):
        cls.image = np.ascontiguousarray(ascent()[:, :511], dtype="f")
        cls.data1d = cls.image[0]
        cls.data2d = cls.image
        cls.data3d = np.tile(cls.image[224:-224, 224:-224], (62, 1, 1))
        cls.kernel1d = gaussian_kernel(1.0)
        cls.kernel2d = np.outer(cls.kernel1d, cls.kernel1d)
        cls.kernel3d = np.multiply.outer(cls.kernel2d, cls.kernel1d)
        cls.ctx = ocl.create_context()
        cls.tol = {
            "1D": 1e-4,
            "2D": 1e-3,
            "3D": 1e-3,
        }

    @classmethod
    def tearDownClass(cls):
        cls.data1d = cls.data2d = cls.data3d = cls.image = None
        cls.kernel1d = cls.kernel2d = cls.kernel3d = None

    @staticmethod
    def compare(arr1, arr2):
        return np.max(np.abs(arr1 - arr2))

    @staticmethod
    def print_err(conv):
        errmsg = str(
            """
            Something wrong with %s
            mode=%s, texture=%s
            """
            % (conv.use_case_desc, conv.mode, conv.use_textures)
        )
        return errmsg

    def instantiate_convol(self, shape, kernel, axes=None):
        if self.mode == "constant":
            if not (self.param["use_textures"]) or (
                self.param["use_textures"]
                and not (check_textures_availability(self.ctx))
            ):
                pytest.skip("mode=constant not implemented without textures")
        C = Convolution(
            shape,
            kernel,
            mode=self.mode,
            ctx=self.ctx,
            axes=axes,
            extra_options={"dont_use_textures": not (self.param["use_textures"])},
        )
        return C

    def get_data_and_kernel(self, test_name):
        dims = {
            "test_1D": (1, 1),
            "test_separable_2D": (2, 1),
            "test_separable_3D": (3, 1),
            "test_nonseparable_2D": (2, 2),
            "test_nonseparable_3D": (3, 3),
        }
        dim_data = {1: self.data1d, 2: self.data2d, 3: self.data3d}
        dim_kernel = {
            1: self.kernel1d,
            2: self.kernel2d,
            3: self.kernel3d,
        }
        dd, kd = dims[test_name]
        return dim_data[dd], dim_kernel[kd]

    def get_reference_function(self, test_name):
        ref_func = {
            "test_1D": lambda x, y: scipy_convolve1d(x, y, mode=self.mode),
            "test_separable_2D": lambda x, y: scipy_convolve1d(
                scipy_convolve1d(x, y, mode=self.mode, axis=1),
                y,
                mode=self.mode,
                axis=0,
            ),
            "test_separable_3D": lambda x, y: scipy_convolve1d(
                scipy_convolve1d(
                    scipy_convolve1d(x, y, mode=self.mode, axis=2),
                    y,
                    mode=self.mode,
                    axis=1,
                ),
                y,
                mode=self.mode,
                axis=0,
            ),
            "test_nonseparable_2D": lambda x, y: scipy_convolve(x, y, mode=self.mode),
            "test_nonseparable_3D": lambda x, y: scipy_convolve(x, y, mode=self.mode),
        }
        return ref_func[test_name]

    def template_test(self, test_name):
        data, kernel = self.get_data_and_kernel(test_name)
        conv = self.instantiate_convol(data.shape, kernel)
        if self.param["input_on_device"]:
            data_ref = parray.to_device(conv.queue, data)
        else:
            data_ref = data
        if self.param["output_on_device"]:
            d_res = parray.empty_like(conv.data_out)
            d_res.fill(0)
            res = d_res
        else:
            res = None
        res = conv(data_ref, output=res)
        if self.param["output_on_device"]:
            res = res.get()
        ref_func = self.get_reference_function(test_name)
        ref = ref_func(data, kernel)
        metric = self.compare(res, ref)
        logger.info("%s: max error = %.2e" % (test_name, metric))
        tol = self.tol[str("%dD" % kernel.ndim)]
        assert metric < tol, self.print_err(conv)


def convolution_data_params():
    boundary_handlings = ["reflect", "nearest", "wrap", "constant"]
    use_textures = [True, False]
    input_on_devices = [True, False]
    output_on_devices = [True, False]
    param_vals = list(
        product(boundary_handlings, use_textures, input_on_devices, output_on_devices)
    )
    params = []
    for boundary_handling, use_texture, input_dev, output_dev in param_vals:
        param={
            "boundary_handling": boundary_handling,
            "input_on_device": input_dev,
            "output_on_device": output_dev,
            "use_textures": use_texture,
        }
        params.append(param)

    return params


@pytest.fixture(scope="module", params=convolution_data_params())
def convolution_data(request):
    """Provide a set of convolution data

    The module scope allows to test each function during a single setup of each
    convolution data
    """
    cdata = None
    try:
        cdata = ConvolutionData(request.param)
        cdata.setUpClass()
        yield cdata
    finally:
        cdata.tearDownClass()


@pytest.mark.skipif(ocl is None, reason="OpenCL is missing")
@pytest.mark.skipif(scipy_convolve is None, reason="scipy is missing")
def test_1D(convolution_data):
    convolution_data.template_test("test_1D")

@pytest.mark.skipif(ocl is None, reason="OpenCL is missing")
@pytest.mark.skipif(scipy_convolve is None, reason="scipy is missing")
def test_separable_2D(convolution_data):
    convolution_data.template_test("test_separable_2D")

@pytest.mark.skipif(ocl is None, reason="OpenCL is missing")
@pytest.mark.skipif(scipy_convolve is None, reason="scipy is missing")
def test_separable_3D(convolution_data):
    convolution_data.template_test("test_separable_3D")

@pytest.mark.skipif(ocl is None, reason="OpenCL is missing")
@pytest.mark.skipif(scipy_convolve is None, reason="scipy is missing")
def test_nonseparable_2D(convolution_data):
    convolution_data.template_test("test_nonseparable_2D")

@pytest.mark.skipif(ocl is None, reason="OpenCL is missing")
@pytest.mark.skipif(scipy_convolve is None, reason="scipy is missing")
def test_nonseparable_3D(convolution_data):
    convolution_data.template_test("test_nonseparable_3D")

@pytest.mark.skipif(ocl is None, reason="OpenCL is missing")
@pytest.mark.skipif(scipy_convolve is None, reason="scipy is missing")
def test_batched_2D(convolution_data):
    """
    Test batched (nonseparable) 2D convolution on 3D data.
    In this test: batch along "z" (axis 0)
    """
    data = convolution_data.data3d
    kernel = convolution_data.kernel2d
    conv = convolution_data.instantiate_convol(data.shape, kernel, axes=(0,))
    res = conv(data)  # 3D
    ref = scipy_convolve(data[0], kernel, mode=convolution_data.mode)  # 2D

    std = np.std(res, axis=0)
    std_max = np.max(np.abs(std))
    assert std_max < convolution_data.tol["2D"], convolution_data.print_err(conv)
    metric = convolution_data.compare(res[0], ref)
    logger.info("test_nonseparable_3D: max error = %.2e" % metric)
    assert metric < convolution_data.tol["2D"], convolution_data.print_err(conv)



#!/usr/bin/env python
# /*##########################################################################
#
# Copyright (c) 2004-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This script is a simple example of how to use the periodic table widgets,
select elements and connect signals.
"""
import sys
from silx.gui import qt
from silx.gui.widgets import PeriodicTable

a = qt.QApplication(sys.argv)
sys.excepthook = qt.exceptionHandler
a.lastWindowClosed.connect(a.quit)

w = qt.QTabWidget()

pt = PeriodicTable.PeriodicTable(w, selectable=True)
pc = PeriodicTable.PeriodicCombo(w)
pl = PeriodicTable.PeriodicList(w)

pt.setSelection(['Fe', 'Si', 'Mt'])
pl.setSelectedElements(['H', 'Be', 'F'])
pc.setSelection("Li")


def change_list(items):
    print("New list selection:", [item.symbol for item in items])


def change_combo(item):
    print("New combo selection:", item.symbol)


def click_table(item):
    print("New table click: %s (%s)" % (item.name, item.subcategory))


def change_table(items):
    print("New table selection:", [item.symbol for item in items])


pt.sigElementClicked.connect(click_table)
pt.sigSelectionChanged.connect(change_table)
pl.sigSelectionChanged.connect(change_list)
pc.sigSelectionChanged.connect(change_combo)

# move combo into container widget to preventing it from filling
# the tab inside TabWidget
comboContainer = qt.QWidget(w)
comboContainer.setLayout(qt.QVBoxLayout())
comboContainer.layout().addWidget(pc)

w.addTab(pt, "PeriodicTable")
w.addTab(pl, "PeriodicList")
w.addTab(comboContainer, "PeriodicCombo")
w.show()

a.exec()



# /*##########################################################################
#
# Copyright (c) 2017-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
This module provides the :class:`SceneWidget` content and parameters model.
"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "11/01/2018"


import weakref

from ... import qt

from .core import BaseRow
from .items import Settings, nodeFromItem


def visitQAbstractItemModel(model, parent=qt.QModelIndex()):
    """Iterate over indices in the model starting from parent

    It iterates column by column and row by row
    (i.e., from left to right and from top to bottom).
    Parent are returned before their children.
    It only iterates through the children for the first column of a row.

    :param QAbstractItemModel model: The model to visit
    :param QModelIndex parent:
        Index from which to start visiting the model.
        Default: start from the root
    """
    assert isinstance(model, qt.QAbstractItemModel)
    assert isinstance(parent, qt.QModelIndex)
    assert parent.model() is model or not parent.isValid()

    for row in range(model.rowCount(parent)):
        for column in range(model.columnCount(parent)):
            index = model.index(row, column, parent)
            yield index

        index = model.index(row, 0, parent)
        for index in visitQAbstractItemModel(model, index):
            yield index


class Root(BaseRow):
    """Root node of :class:`SceneWidget` parameters.

    It has two children:
    - Settings
    - Scene group
    """

    def __init__(self, model, sceneWidget):
        super(Root, self).__init__()
        self._sceneWidget = weakref.ref(sceneWidget)
        self.setParent(model)  # Needed for Root

    def children(self):
        sceneWidget = self._sceneWidget()
        if sceneWidget is None:
            return ()
        else:
            return super(Root, self).children()


class SceneModel(qt.QAbstractItemModel):
    """Model of a :class:`SceneWidget`.

    :param SceneWidget parent: The SceneWidget this model represents.
    """

    def __init__(self, parent):
        self._sceneWidget = weakref.ref(parent)

        super(SceneModel, self).__init__(parent)
        self._root = Root(self, parent)
        self._root.addRow(Settings(parent))
        self._root.addRow(nodeFromItem(parent.getSceneGroup()))

    def sceneWidget(self):
        """Returns the :class:`SceneWidget` this model represents.

        In case the widget has already been deleted, it returns None

        :rtype: SceneWidget
        """
        return self._sceneWidget()

    def _itemFromIndex(self, index):
        """Returns the corresponding :class:`Node` or :class:`Item3D`.

        :param QModelIndex index:
        :rtype: Node or Item3D
        """
        return index.internalPointer() if index.isValid() else self._root

    def index(self, row, column, parent=qt.QModelIndex()):
        """See :meth:`QAbstractItemModel.index`"""
        if column >= self.columnCount(parent) or row >= self.rowCount(parent):
            return qt.QModelIndex()

        item = self._itemFromIndex(parent)
        return self.createIndex(row, column, item.children()[row])

    def parent(self, index):
        """See :meth:`QAbstractItemModel.parent`"""
        if not index.isValid():
            return qt.QModelIndex()

        item = self._itemFromIndex(index)
        parent = item.parent()

        ancestor = parent.parent()

        if ancestor is not self:  # root node
            children = ancestor.children()
            row = children.index(parent)
            return self.createIndex(row, 0, parent)

        return qt.QModelIndex()

    def rowCount(self, parent=qt.QModelIndex()):
        """See :meth:`QAbstractItemModel.rowCount`"""
        item = self._itemFromIndex(parent)
        return item.rowCount()

    def columnCount(self, parent=qt.QModelIndex()):
        """See :meth:`QAbstractItemModel.columnCount`"""
        item = self._itemFromIndex(parent)
        return item.columnCount()

    def data(self, index, role=qt.Qt.DisplayRole):
        """See :meth:`QAbstractItemModel.data`"""
        item = self._itemFromIndex(index)
        column = index.column()
        return item.data(column, role)

    def setData(self, index, value, role=qt.Qt.EditRole):
        """See :meth:`QAbstractItemModel.setData`"""
        item = self._itemFromIndex(index)
        column = index.column()
        if item.setData(column, value, role):
            self.dataChanged.emit(index, index)
            return True
        return False

    def flags(self, index):
        """See :meth:`QAbstractItemModel.flags`"""
        item = self._itemFromIndex(index)
        column = index.column()
        return item.flags(column)

    def headerData(self, section, orientation, role=qt.Qt.DisplayRole):
        """See :meth:`QAbstractItemModel.headerData`"""
        if orientation == qt.Qt.Horizontal and role == qt.Qt.DisplayRole:
            return 'Item' if section == 0 else 'Value'
        else:
            return None



# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/

__authors__ = ["P. Knobel"]
__license__ = "MIT"
__date__ = "17/03/2016"

cimport cython

cdef extern from "SpecFileCython.h":
    struct _SpecFile:
        pass
# Renaming struct because we have too many SpecFile items (files, classes…)
ctypedef _SpecFile SpecFileHandle

cdef extern from "SpecFileCython.h":
    # sfinit
    SpecFileHandle* SfOpen(char*, int*)
    int SfClose(SpecFileHandle*)
    char* SfError(int)
    
    # sfindex
    long* SfList(SpecFileHandle*, int*)
    long SfScanNo(SpecFileHandle*)
    long SfIndex(SpecFileHandle*, long, long)
    long SfNumber(SpecFileHandle*, long)
    long SfOrder(SpecFileHandle*, long)
    
    # sfdata
    int SfData(SpecFileHandle*, long, double***, long**, int*)
    long SfDataLine(SpecFileHandle*, long, long, double**, int*)
    long SfDataColByName(SpecFileHandle*, long, char*, double**, int*)
    
    # sfheader
    #char* SfTitle(SpecFileHandle*, long, int*)
    long SfHeader(SpecFileHandle*, long, char*, char***, int*)
    long SfFileHeader(SpecFileHandle*, long, char*, char***, int*)
    char* SfCommand(SpecFileHandle*, long, int*)
    long SfNoColumns(SpecFileHandle*, long, int*)
    char* SfDate(SpecFileHandle*, long, int*)
    
    # sflabel
    long SfAllLabels(SpecFileHandle*, long, char***, int*)
    char* SfLabel(SpecFileHandle*, long, long, int *)
    long SfAllMotors(SpecFileHandle*, long, char***, int*)
    long SfAllMotorPos(SpecFileHandle*, long, double**, int*)
    double SfMotorPosByName(SpecFileHandle*, long, char*, int*)
    
    # sftools
    void freeArrNZ(void***, long)

    # sfmca
    long SfNoMca(SpecFileHandle*, long, int*)
    int  SfGetMca(SpecFileHandle*, long, long , double**, int*)
    long SfMcaCalib(SpecFileHandle*, long, double**, int*)




# /*##########################################################################
#
# Copyright (c) 2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module provides a widget to visualize 3D arrays"""

__authors__ = ["T. Vincent"]
__license__ = "MIT"
__date__ = "22/03/2019"


import numpy

from .. import qt
from ..plot3d.SceneWindow import SceneWindow
from ..plot3d.items import ScalarField3D, ComplexField3D, ItemChangedType


class VolumeWindow(SceneWindow):
    """Extends SceneWindow with a convenient API for 3D array

    :param QWidget: parent
    """

    def __init__(self, parent):
        super(VolumeWindow, self).__init__(parent)
        self.__firstData = True
        # Hide global parameter dock
        self.getGroupResetWidget().parent().setVisible(False)

    def setAxesLabels(self, xlabel=None, ylabel=None, zlabel=None):
        """Set the text labels of the axes.

        :param Union[str,None] xlabel: Label of the X axis
        :param Union[str,None] ylabel: Label of the Y axis
        :param Union[str,None] zlabel: Label of the Z axis
        """
        sceneWidget = self.getSceneWidget()
        sceneWidget.getSceneGroup().setAxesLabels(
            'X' if xlabel is None else xlabel,
            'Y' if ylabel is None else ylabel,
            'Z' if zlabel is None else zlabel)

    def clear(self):
        """Clear any currently displayed data"""
        sceneWidget = self.getSceneWidget()
        items = sceneWidget.getItems()
        if (len(items) == 1 and
                isinstance(items[0], (ScalarField3D, ComplexField3D))):
            items[0].setData(None)
        else:  # Safety net
            sceneWidget.clearItems()

    @staticmethod
    def __computeIsolevel(data):
        """Returns a suitable isolevel value for data

        :param numpy.ndarray data:
        :rtype: float
        """
        data = data[numpy.isfinite(data)]
        if len(data) == 0:
            return 0
        else:
            return numpy.mean(data) + numpy.std(data)

    def setData(self, data, offset=(0., 0., 0.), scale=(1., 1., 1.)):
        """Set the 3D array data to display.

        :param numpy.ndarray data: 3D array of float or complex
        :param List[float] offset: (tx, ty, tz) coordinates of the origin
        :param List[float] scale: (sx, sy, sz) scale for each dimension
        """
        sceneWidget = self.getSceneWidget()
        dataMaxCoords = numpy.array(list(reversed(data.shape))) - 1

        previousItems = sceneWidget.getItems()
        if (len(previousItems) == 1 and
                isinstance(previousItems[0], (ScalarField3D, ComplexField3D)) and
                numpy.iscomplexobj(data) == isinstance(previousItems[0], ComplexField3D)):
            # Reuse existing volume item
            volume = sceneWidget.getItems()[0]
            volume.setData(data, copy=False)
            # Make sure the plane goes through the dataset
            for plane in volume.getCutPlanes():
                point = numpy.array(plane.getPoint())
                if numpy.any(point < (0, 0, 0)) or numpy.any(point > dataMaxCoords):
                    plane.setPoint(dataMaxCoords // 2)
        else:
            # Add a new volume
            sceneWidget.clearItems()
            volume = sceneWidget.addVolume(data, copy=False)
            volume.setLabel('Volume')
            for plane in volume.getCutPlanes():
                # Make plane going through the center of the data
                plane.setPoint(dataMaxCoords // 2)
                plane.setVisible(False)
                plane.sigItemChanged.connect(self.__cutPlaneUpdated)
            volume.addIsosurface(self.__computeIsolevel, '#FF0000FF')

            # Expand the parameter tree
            model = self.getParamTreeView().model()
            index = qt.QModelIndex()  # Invalid index for top level
            while 1:
                rowCount = model.rowCount(parent=index)
                if rowCount == 0:
                    break
                index = model.index(rowCount - 1, 0, parent=index)
                self.getParamTreeView().setExpanded(index, True)
                if not index.isValid():
                    break

        volume.setTranslation(*offset)
        volume.setScale(*scale)

        if self.__firstData:  # Only center for first dataset
            self.__firstData = False
            sceneWidget.centerScene()

    def __cutPlaneUpdated(self, event):
        """Handle the change of visibility of the cut plane

        :param event: Kind of update
        """
        if event == ItemChangedType.VISIBLE:
            plane = self.sender()
            if plane.isVisible():
                self.getSceneWidget().selection().setCurrentItem(plane)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""PyQt6 backward compatibility patching"""

__authors__ = ["Thomas VINCENT"]
__license__ = "MIT"
__date__ = "02/09/2021"

import enum
import logging

import PyQt6.sip
from PyQt6.QtCore import Qt


_logger = logging.getLogger(__name__)


def patch_enums(*modules):
    """Patch PyQt6 modules to provide backward compatibility of enum values

    :param modules: Modules to patch (e.g., PyQt6.QtCore).
    """
    for module in modules:
        for clsName in dir(module):
            cls = getattr(module, clsName, None)
            if isinstance(cls, PyQt6.sip.wrappertype) and clsName.startswith('Q'):
                for qenumName in dir(cls):
                    if qenumName[0].isupper():
                        qenum = getattr(cls, qenumName, None)
                        if isinstance(qenum, enum.EnumMeta):
                            if qenum is getattr(cls.__mro__[1], qenumName, None):
                                continue  # Only handle it once
                            for item in qenum:
                                # Special cases to avoid overrides and mimic PySide6
                                if clsName == 'QColorSpace' and qenumName in (
                                        'Primaries', 'TransferFunction'):
                                    break
                                if qenumName in ('DeviceType', 'PointerType'):
                                    break

                                setattr(cls, item.name, item)
